"""
This class manages running the checker against a list of functions.
There is also support for low level progress for functions/classes.
"""
import datetime as dt
from abc import ABC, abstractmethod
from typing import Any, Sequence

from .ten8t_exception import Ten8tException
from .ten8t_format import Ten8tAbstractRender, Ten8tRenderText
from .ten8t_function import Ten8tFunction
from .ten8t_immutable import Ten8tEnvDict, Ten8tEnvList, Ten8tEnvSet
from .ten8t_logging import ten8t_logger
from .ten8t_module import Ten8tModule
from .ten8t_package import Ten8tPackage
from .ten8t_rc import Ten8tRC
from .ten8t_result import Ten8tResult
from .ten8t_ruid import empty_ruids, ruid_issues, valid_ruids
from .ten8t_score import ScoreByResult, ScoreStrategy


# pylint: disable=R0903
class Ten8tProgress(ABC):
    """ Base class for all ten8t progress bars"""

    def __init__(self):
        pass

    @abstractmethod
    def __call__(self,
                 current_iteration: int,
                 max_iterations,
                 text: str,
                 result=None):  # pragma: no cover
        pass


# pylint: disable=R0903
class Ten8tNoProgress(Ten8tProgress):
    """Progress 'bar' that does nothing.  Useful for testing and debugging."""

    def __call__(self, current_iteration: int,
                 max_iterations,
                 text: str, result=None):
        """Don't do anything for progress.  This is useful for testing."""


# pylint: disable=R0903
class Ten8tDebugProgress(Ten8tProgress):
    """ Progress 'bar' that is useful for debugging by printing data to the console"""

    def __call__(self, current_iteration: int, max_iteration: int, msg: str,
                 result=None):  # pylint: disable=unused-argument
        """Print a debug message."""
        if msg:
            print(msg)
        if result:
            print("+" if result.status else "-", end="")


def _param_str_list(params: Sequence[str] | str | None,
                    disallowed=' ,!@#$%^&*(){}[]<>~`-+=\t\n\'"') -> Sequence[str]:
    """
    Allow user to specify "foo fum" instead of ["foo","fum"] or slightly more
    shady "foo" instead of ["foo"].  This is strictly for reducing friction
    for the programmer.

    Also note: the disallowed characters is just me being courteous and trying to protect
               you from yourself.  If there are other dumb characters that
               I missed please submit a PR as I have no intention of walling off
               everything in a dynamic language.

    Returns: List of Strings

    Args:
        params: "foo fum" or ["foo","fum"]

    """

    # Null case...on could argue they meant the empty string as a name
    if params is None or params == [] or isinstance(params, str) and params.strip() == '':
        return []

    if isinstance(params, str):
        params = params.split()

    for param in params:
        if not isinstance(param, str):
            raise Ten8tException(f"Invalid parameter list {param}")
        bad_chars = [c for c in disallowed if c in param]
        if bad_chars:
            raise Ten8tException(f"Parameter '{bad_chars}' has a space in it.  ")

    return params


def _param_int_list(params: list[str] | list[int] | int | str | None) -> list[int]:
    """
    That's a lot of options there.
    
    Allow user to specify "1 2 3" instead of [1,2,3] or slightly more
    shady 1 instead of [1].  For small numbers this is a wash but for
    symmetry with str_list it included it.

    NOTE: The separator is the default for split...whitespace

    Args:
        params: "1 2" or [1,2] or even [1,"2"]

    Returns: List of Integers

    """

    if isinstance(params, int):
        return [params]

    elif isinstance(params, str):
        params = params.split()

    params = [str(p) for p in params]

    # Make sure everything is an integer
    for param in params:
        if isinstance(param, str) and param.isdigit():
            continue
        raise Ten8tException(f"Invalid integer parameter in {param} in {params}")

    return [int(param) for param in params]


def exclude_ruids(ruids: list[str]):
    """Return a filter function that will exclude the ruids from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.ruid not in ruids

    return filter_func


def exclude_tags(tags: list[str]):
    """Return a filter function that will exclude the tags from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.tag not in tags

    return filter_func


def exclude_levels(levels: list[int]):
    """Return a filter function that will exclude the levels from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.level not in levels

    return filter_func


def exclude_phases(phases: list[str]):
    """Return a filter function that will exclude the phases from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.phase not in phases

    return filter_func


def keep_ruids(ruids: list[str]):
    """Return a filter function that will keep the ruids from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.ruid in ruids

    return filter_func


def keep_tags(tags: list[str]):
    """Return a filter function that will keep the tags from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.tag in tags

    return filter_func


def keep_levels(levels: list[int]):
    """Return a filter function that will keep the levels from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.level in levels

    return filter_func


def keep_phases(phases: list[str]):
    """Return a filter function that will keep the phases from the list."""

    def filter_func(s_func: Ten8tFunction):
        return s_func.phase in phases

    return filter_func


def debug_progress(_, msg: str | None = None, result: Ten8tResult | None = None
                   ):  # pylint: disable=unused-argument
    """Print a debug message."""
    if msg:
        print(msg)
    if result:
        print("+" if result.status else "-", end="")


class Ten8tChecker:
    """
    A checker object is what manages running rules against a system.

    THe life cycle of a checker object is

    1) Load what ever packages/modules/functions are associated with a system as
       a collection of functions that could be run.
    2) Load any environment that may be needed for the rules.
    2) Optionally filter those functions based on any of the function attributes.
    3) Check all the rules and collect the results while providing status using
       a user specified progress object.
    4) Score the results based on the scoring strategy.
    5) Return the result object as object data or json data.
    """

    def __init__(
            self,
            packages: list[Ten8tPackage] | None = None,
            modules: list[Ten8tModule] | None = None,
            check_functions: list[Ten8tFunction] | None = None,
            progress_object: Ten8tProgress | None = None,
            score_strategy: ScoreStrategy | None = None,
            rc: Ten8tRC | None = None,
            env: dict[str, Any] | None = None,
            renderer: Ten8tAbstractRender = None,
            abort_on_fail=False,
            abort_on_exception=False,
            auto_setup: bool = False,
            auto_ruid: bool = False,
            auto_thread: str | bool = None,
    ):
        """

        
        Args:
            packages: List of Ten8tPackage objs to check. 
                      If not provided, default = [] .
            modules: A list of Ten8tModule objs to check. 
                     If not provided, default = [].
            check_functions: A list of Ten8tFunction objs to check. If not provided, default = [].
            progress_object: A Ten8tProgress objs for tracking progress. 
                             If not provided, def = Ten8tNoProgress.
            score_strategy: A ScoreStrategy objs for scoring the results. 
                            If not provided, def = ScoreByResult.
            env: A dict containing additional env variables. 
                            If not provided, def = {}
            abort_on_fail: A bool flag indicating whether to abort a fail result occurs. def =False.
            abort_on_exception: A bool flag indicating whether to abort on exceptions. def=False.
            auto_setup: A bool flag automatically invoke pre_collect/prepare. def=False.
            auto_ruid: A bool flag automatically generate rule_ids if they don't exist.
            auto_thread: A bool flag to indicate if the check functions should be automatically put in threads.
        Raises:
            Ten8tException: If the provided packages, modules, or check_functions 
                             are not in the correct format.

        """

        self.packages = self._process_packages(packages)
        self.modules = self._process_modules(modules)
        self.check_functions = self._process_check_funcs(check_functions)

        # If the user has not provided a score strategy then use the simple one
        self.score_strategy = score_strategy or ScoreByResult()
        self.score = 0.0

        # Allow an RC object to be specified.
        self.rc = rc

        # In order to support rendered output a render object must be provided
        # if none are provided we create one
        self.renderer = renderer or Ten8tRenderText()

        # If we are provided with an environment we save it off but first wrap it in
        # a class that guards reasonably against writes to the underlying environment
        # data.
        if env:
            self.env = self._make_immutable_env(env)
        else:
            self.env = {}

        # THis dict has the environment values that are NULL
        self.env_nulls: dict[str, Any] = {}

        # Connect the progress output to the checker object.  The NoProgress
        # class is a dummy class that does no progress reporting.
        self.progress_callback: Ten8tProgress = progress_object or Ten8tNoProgress()

        # If any fail result occurs stop processing.
        self.abort_on_fail = abort_on_fail

        # If any exception occurs stop processing
        self.abort_on_exception = abort_on_exception

        # These two have the collection of all checker functions from packages, modules, and adhoc
        self.collected: list[Ten8tFunction] = []
        self.pre_collected: list[Ten8tFunction] = []

        self.start_time = dt.datetime.now()
        self.end_time = dt.datetime.now()
        self.results: list[Ten8tResult] = []
        self.auto_ruid = auto_ruid

        if not self.packages and not self.modules and not self.check_functions:
            raise Ten8tException(
                "You must provide at least one package, module or function to check."
            )

        # For some use cases there is no need for special setup so just do auto setup
        # to clean up the startup.  Real code will likely need to be sophisticated
        # with prepare...
        if auto_setup:
            self.pre_collect()
            self.prepare()

        # Useful in some contexts.  Instead of manually setting these
        self._set_auto_thread_names(auto_thread)

    def _set_auto_thread_names(self, auto_thread: str):
        """
        Assigns thread names automatically to functions in `self.collected` if they don't already
        have a `thread_id`. Uses a naming pattern with placeholders for indexing and RUIDs.

        Args:
            auto_thread (str): The thread naming template. If `True`, defaults to "auto_{i}_{ruid}".
        """
        # Default naming template if auto_thread is True
        if not auto_thread:
            return
        if auto_thread is True:
            auto_thread = "auto_{i}_{ruid}_autothread"

        for i, func in enumerate(self.collected):
            if not func.thread_id:
                # Use `str.format()` to replace placeholders with values
                func.thread_id = auto_thread.format(i=i, ruid=func.ruid, name=func.function_name, module=func.module)

    @staticmethod
    def _make_immutable_env(env: dict) -> dict:
        """
        Converts mutable containers in a dictionary to immutable versions.
        """
        for key, value in env.items():

            # Detect mutable objects and convert them to immutable ones
            if isinstance(value, list):
                env[key] = Ten8tEnvList(value)
            elif isinstance(value, dict):
                env[key] = Ten8tEnvDict(value)
            elif isinstance(value, set):
                env[key] = Ten8tEnvSet(value)

        return env

    @staticmethod
    def _process_packages(packages: list[Ten8tPackage] | None) -> list[Ten8tPackage]:
        """ Allow packages to be in various forms"""
        if not packages:
            return []
        if isinstance(packages, Ten8tPackage):
            return [packages]
        if isinstance(packages, list) and all(isinstance(p, Ten8tPackage) for p in packages):
            return packages
        raise Ten8tException('Packages must be a list of Ten8tPackage objects.')

    @staticmethod
    def _process_modules(modules: list[Ten8tModule] | None) -> list[Ten8tModule]:
        """ Allow modules to be in various forms"""
        if not modules:
            return []
        if isinstance(modules, Ten8tModule):
            return [modules]
        if isinstance(modules, list) and all(isinstance(m, Ten8tModule) for m in modules):
            return modules
        raise Ten8tException('Modules must be a list of Ten8tModule objects.')

    @staticmethod
    def _process_check_funcs(check_functions: list[Ten8tFunction] | None) -> list[Ten8tFunction]:
        """ Load up an arbitrary list of ten8t functions.
        These functions are tagged with adhoc for module"""
        if isinstance(check_functions, list) and len(check_functions) >= 1:
            for count, f in enumerate(check_functions, start=1):
                if not isinstance(f, Ten8tFunction):
                    raise Ten8tException(
                        "Functions must be a list of Ten8tFunction objects."
                    )
                # Since we are building up a module from nothing we give it a generic name and
                # remember the load order.
                f.index = count
                f.module = "adhoc"
            return check_functions

        if not check_functions:
            return []

        raise Ten8tException("Functions must be a list of Ten8tFunction objects.")

    def pre_collect(self) -> list[Ten8tFunction]:
        """
        Collect all the functions from the packages, modules and functions with no filtering.
        This list of functions is will be filtered by the checker before running checks.

        Returns:
            _type_: _description_
        """

        self.pre_collected = []

        # Worth noting that += does an 'extend' rather than an 'append'.
        for pkg in self.packages:
            self.pre_collected += [func for module in pkg.modules for func in module.check_functions]

        for module in self.modules:
            self.pre_collected += [func for func in module.check_functions]

        self.pre_collected += [func for func in self.check_functions]

        # List of all possible functions that could be run
        return self.pre_collected

    def prepare(self, filter_functions=None):
        """
        Prepare the collected functions for running checks.

        Run through the collected functions to prepare the checks that will be run.
        A list of filter functions may be provided to filter the functions. Filter
        functions must return True if the function should be kept.

        Args:
            filter_functions (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """

        # If no filter functions are provided then use a default one allows all functions
        filter_functions = filter_functions or [lambda _: True]

        self.auto_gen_ruids()

        # At this point we have all the functions in the packages, modules and functions
        # Now we need to filter out the ones that are not wanted. Filter functions return
        # True if the function should be kept
        self.collected = []
        for ten8t_func in self.pre_collected:
            if all(f(ten8t_func) for f in filter_functions):
                self.collected.append(ten8t_func)

        # Now use the RC file.  Note that if you are running filter functions AND
        # an RC file this can be confusing.  Ideally you use one or the other. but
        # it isn't an error to do so, you just need to know what you are doing.
        self.apply_rc(self.rc)

        # The collected list has the functions that will be run to verify operation
        # of the system.

        # If the user has provided valid ruids for all functions (or for none) then
        # we can proceed.  If not then we need to raise an exception and show the issues.
        ruids = [f.ruid for f in self.collected]

        # If the user decided to set up ruids for every function OR if they didn't configure
        # any ruids then we can just run with the collected functions.
        if empty_ruids(ruids) or valid_ruids(ruids):
            return self.collected

        # Otherwise there is a problem.
        raise Ten8tException(
            f"There are duplicate or missing RUIDS: {ruid_issues(ruids)}"
        )

    def auto_gen_ruids(self, template='__ruid__@id@'):
        """ Provide a mechanism for to transition from no ruids to ruids.  This way they
            can only set up the rules that need rule_ids"""
        if not self.auto_ruid:
            return
        id_ = 1
        for function in self.pre_collected:
            if function.ruid == '':
                function.ruid = template.replace("@id@", f'{id_:04d}')
                id_ += 1

    def apply_rc(self, rc=None):
        """ Apply RC file to collected functions applying includes then excludes. """
        self.rc = rc or self.rc

        # By exiting early an NOT using the RC file we don't use
        # Sets as shown below.  Sets cause order to be nondeterministic
        if not self.rc:
            return self.collected

        self.collected = [function for function in self.collected
                          if self.rc.does_match(ruid=function.ruid,
                                                tag=function.tag,
                                                phase=function.phase,
                                                level=function.level)]

        return self.collected

    def exclude_by_attribute(self, tags: Sequence[str] | str | None = None,
                             ruids: Sequence[str] | str | None = None,
                             levels: Sequence[int] | int | None = None,
                             phases: Sequence[str] | str | None = None) -> list[Ten8tFunction]:
        """ Run everything except the ones that match these attributes """

        # Make everything nice lists
        tags = _param_str_list(tags)
        ruids = _param_str_list(ruids)
        phases = _param_str_list(phases)
        levels = _param_int_list(levels)

        # Exclude attributes that don't match
        self.collected = [f for f in self.collected if f.tag not in tags and
                          f.ruid not in ruids and
                          f.level not in levels and
                          f.phase not in phases]
        return self.collected

    def include_by_attribute(self,
                             tags: list | str | None = None,
                             ruids: list | str | None = None,
                             levels: list | str | None = None,
                             phases: list | str | None = None) -> list[Ten8tFunction]:
        """ Run everything that matches these attributes """

        # Make everything nice lists
        tags_ = _param_str_list(tags)
        ruids_ = _param_str_list(ruids)
        phases_ = _param_str_list(phases)
        levels_ = _param_int_list(levels)

        # This is a special case to make including everything the default
        if not tags and not ruids and not levels and not phases:
            return self.collected

        # Only include the attributes that match
        self.collected = [f for f in self.collected if (f.tag in tags_) or
                          (f.ruid in ruids_) or
                          (f.level in levels_) or
                          (f.phase in phases_)]

        return self.collected

    def load_environments(self):
        """
        This takes the global environment and adds in the results
        from all the discovered environment functions.  The results
        are all merged into a dictionary of parameter names and their values.

        This works very much like pytest, only without the scoping Parameters
        that are needed in multiple places aren't regenerated.
        Returns:

        """

        # Prime the environment with top level config
        # This should be json-able things
        full_env = self.env.copy()

        for m in self.modules:
            for env_func in m.env_functions:
                # TODO: There should be exceptions on collisions
                full_env.update(env_func(full_env))

        # This is a concern, there should be no nulls, HOWEVER this is more complex
        # since there should be no nulls for parameters to the collected check functions.
        # for now, I'm tracking this and dumping it in the results.
        self.env_nulls = [key for key, value in full_env.items() if value is None]

        return full_env

    @property
    def ruids(self):
        """
        Return a list of all the RUIDs in the collected functions.

        Returns:
            _type_: _description_
        """
        r = sorted(set(f.ruid for f in self.collected))
        return r

    @property
    def levels(self):
        """
        Return a list of all the levels in the collected functions.

        Returns:
            _type_: _description_
        """
        return sorted(set(f.level for f in self.collected))

    @property
    def tags(self):
        """
        Return a list of all the tags in the collected functions.

        Returns:
            _type_: _description_
        """
        return sorted(set(f.tag for f in self.collected))

    @property
    def phases(self):
        """
        Return a list of all the phases in the collected functions.

        Returns:
            _type_: _description_
        """
        return sorted(set(f.phase for f in self.collected))

    class AbortYieldException(Exception):
        """Allow breaking out of multi level loop without state variables"""

    def yield_all(self, env=None):
        """
        Yield all the results from the collected functions

        This is where the rule engine does its work.

        Args:
            env: The environment to use for the rule functions

        Yields:
            _type_: Ten8tResult
        """

        # Note that it is possible for the collected list to be
        # empty.  This is not an error condition.  It is possible
        # that the filter functions have filtered out all the
        # functions.
        count = 0
        self.progress_callback(count, self.function_count, "Start Rule Check")
        self.start_time = dt.datetime.now()

        ten8t_logger.info("Checker start with %d functions", len(self.collected))

        try:
            # Magic happens here.  Each module is checked for any functions that start with 
            # env_ (which is configurable).  Env is a dictionary that has values that may be
            # used as function parameters to check functions (very similar to pytest).  At this
            # time environments are global, hence there could be collisions on larger projects.
            env = self.load_environments()

            # Shuts up linter
            # result: Ten8tResult | None = None
            # function: Ten8tFunction | None = None

            # Count here to enable progress bars
            for count, function_ in enumerate(self.collected, start=1):

                # Lots of magic here
                function_.env = env

                self.progress_callback(count,
                                       self.function_count,
                                       f"Func Start {function_.function_name}")
                for result in function_():

                    # Render the message if needed.  The render happens right before it is yielded so it "knows" as 
                    # much as possible at this point.
                    result.msg_rendered = result.msg if not self.renderer else self.renderer.render(result.msg)

                    ten8t_logger.debug("%s:%s:%s", result.func_name, result.status, result.msg)

                    yield result

                    # Check early exits
                    if self.abort_on_fail and result.status is False:
                        ten8t_logger.info("Abort on fail")
                        raise self.AbortYieldException()

                    if self.abort_on_exception and result.except_:
                        ten8t_logger.info("Abort on exception")
                        raise self.AbortYieldException()

                    # Stop yielding from a function
                    if function_.finish_on_fail and result.status is False:
                        self.progress_callback(count, self.function_count,
                                               f"Early exit. {function_.function_name} failed.")
                        break
                    self.progress_callback(count, self.function_count, "", result)
                self.progress_callback(count, self.function_count, "Func done.")

        except self.AbortYieldException:
            name = function_.function_name if function_ is not None else "???"

            if self.abort_on_fail:
                self.progress_callback(count,
                                       self.function_count,
                                       f"Abort on fail: {name}")
            if self.abort_on_exception:
                self.progress_callback(count,
                                       self.function_count,
                                       f"Abort on exception: {name}")

        self.end_time = dt.datetime.now()
        self.progress_callback(count,
                               self.function_count,
                               "Rule Check Complete.")
        ten8t_logger.info("Checker complete ran %s check functions", self.function_count)


    def run_all(self, env=None) -> list[Ten8tResult]:
        """
        List version of yield all.

        """

        # A deceptively important line of code
        self.results = list(self.yield_all(env=env))

        self.score = self.score_strategy(self.results)
        self.progress_callback(self.function_count,
                               self.function_count,
                               f"Score = {self.score:.1f}")
        return self.results

    @property
    def clean_run(self):
        """ No exceptions """
        return all(not r.except_ for r in self.results)

    @property
    def perfect_run(self):
        """No fails or skips"""
        return all(r.status and not r.skipped and not r.warn_msg for r in self.results)

    @property
    def skip_count(self):
        """Number of skips"""
        return len([r for r in self.results if r.skipped])

    @property
    def warn_count(self):
        """Number of warns"""
        return len([r for r in self.results if r.warn_msg])

    @property
    def pass_count(self):
        """Number of passes"""
        return len([r for r in self.results if r.status and not r.skipped])

    @property
    def fail_count(self):
        """Number of fails"""
        return len([r for r in self.results if not r.status and not r.skipped])

    @property
    def result_count(self):
        """ Possibly redundant call to get the number of results."""
        return len(self.results)

    @property
    def function_count(self):
        """ Count all the collected functions (this is not the ones that are run)"""
        return len(self.collected)

    @property
    def module_count(self):
        """Count up all the modules in all the packages"""
        return len(self.modules) + sum(1 for pkg in self.packages for _ in pkg.modules)

    @property
    def module_names(self):
        """Get a list of module names."""
        return [module.name for module in self.modules] + \
            [m.module_name for pkg in self.packages for m in pkg.modules]

    @property
    def package_count(self):
        """ Count of the packages (almost always 1 or 0)"""
        return len(self.packages)

    def get_header(self) -> dict:
        """Make a header with the top level information about the checker run"""
        header = {
            "package_count": self.package_count,
            "module_count": self.module_count,
            "modules": self.module_names,
            "function_count": self.function_count,
            "tags": self.tags,
            "levels": self.levels,
            "phases": self.phases,
            "ruids": self.ruids,
            "score": self.score,
            "env_nulls": self.env_nulls,
        }
        return header

    def as_dict(self):
        """
        Return a dictionary of the results.
        """
        h = self.get_header()

        r = {
            "start_time": self.start_time,
            "end_time": self.end_time,
            "duration_seconds": (self.end_time - self.start_time).total_seconds(),

            "functions": [f.function_name for f in self.check_functions],
            "passed_count": self.pass_count,
            "failed_count": self.fail_count,
            "skip_count": self.skip_count,
            "total_count": self.result_count,

            # the meat of the output lives here
            "results": [r.as_dict() for r in self.results],
        }
        return h | r



"""
This module makes dealing with configuration files a bit easier as it supports JSON and TOML
out of the box.

I also have begun using the patch statement and this was a reasonable place to use it soe
I have provided two implementations.
"""

import sys

from .ten8t_exception import Ten8tException
from .ten8t_inirc import Ten8tIniRC
from .ten8t_jsonrc import Ten8tJsonRC
from .ten8t_rc import Ten8tRC
from .ten8t_tomlrc import Ten8tTomlRC
from .ten8t_xmlrc import Ten8tXMLRC

if sys.version_info[:2] >= (3, 10):
    def ten8t_rc_factory(param: dict | str, section: str = "") -> Ten8tRC:
        """
        Factory function to create an instance of Ten8tRC or its subclasses for Python 3.10 and above.
        """
        match param:
            case dict(d):
                if section == "":
                    return Ten8tRC(rc_d=d)
                else:
                    return Ten8tRC(rc_d=d[section])
            case str(s) if s.endswith('.toml'):
                return Ten8tTomlRC(cfg=s, section=section)
            case str(s) if s.endswith('.json'):
                return Ten8tJsonRC(cfg=s, section=section)
            case str(s) if s.endswith('.xml'):
                return Ten8tXMLRC(cfg=s, section=section)
            case str(s) if s.endswith('.ini'):
                return Ten8tIniRC(cfg=s, section=section)
            case _:
                raise Ten8tException('Invalid parameter type for ten8t_rc_factory.')
else:  # pragma: no cover
    def ten8t_rc_factory(param, section: str = "") -> Ten8tRC:
        """
        Factory function to create an instance of Ten8tRC or its subclasses for Python below 3.10.
        """
        if isinstance(param, dict):
            if not section:
                return Ten8tRC(rc_d=param)
            else:
                return Ten8tRC(rc_d=param[section])
        elif isinstance(param, str):
            if param.endswith('.toml'):
                return Ten8tTomlRC(cfg=param, section=section)
            elif param.endswith('.json'):
                return Ten8tJsonRC(cfg=param, section=section)
            elif param.endswith('.xml'):
                return Ten8tXMLRC(cfg=param, section=section)
            elif param.endswith('.ini'):
                return Ten8tIniRC(cfg=param, section=section)

        raise Ten8tException(f'Invalid parameter type for ten8t_rc_factory {param=}-{section=}.')



"""Baseline setup file """

# pragma no cover
from setuptools import find_packages, setup  # type: ignore

# Mostly AI generated setup.
setup(
    name='ten8t',
    version='0.1.0',
    description='A tool for checking rules against filesystems, databases, and APIs',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    author='Chuck Bass',
    author_email='chuck@acrocad.net',
    url='https://github.com/hucker/ten8t',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'ten8tit=ten8t.ten8t_it:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    python_requires='>=3.8',
)



import logging

# Create a library-specific logger
ten8t_logger = logging.getLogger("ten8t")
ten8t_logger.addHandler(logging.NullHandler())  # Default to NullHandler


def ten8t_setup_logging(
        level: int = logging.WARNING,
        propagate: bool = True,
        file_name: str | None = None,
        stream=None,
        format_string: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
) -> None:
    """
    Configure the global logger for the 'ten8t' package.

    Args:
        level (int): Logging level (default=logging.WARNING).
        propagate (bool): If True, propagates logs to the parent logger (default=True).
        file_name (str): Optional file path to save logs.
        stream (io.TextIOWrapper | None): Optional stream (e.g., sys.stdout, sys.stderr).
        format_string (str): Format string for log messages (default includes timestamp, name, level, and message).

    Returns:
        None: Configures the global `ten8t_logger`.
    """
    global ten8t_logger

    # Create a formatter
    formatter = logging.Formatter(format_string)

    # Add a file handler if file_name is provided
    if file_name:
        try:
            file_handler = logging.FileHandler(file_name)
            file_handler.setFormatter(formatter)
            ten8t_logger.addHandler(file_handler)
        except (OSError, PermissionError) as e:
            raise ValueError(f"Error setting up file handler for '{file_name}'. Details: {e}")

    # Add a stream handler if stream is specified
    if stream:
        if not hasattr(stream, "write"):
            raise ValueError(f"The provided stream '{stream}' is not a valid writable stream.")
        stream_handler = logging.StreamHandler(stream)
        stream_handler.setFormatter(formatter)
        ten8t_logger.addHandler(stream_handler)

    # Configure logger settings
    ten8t_logger.setLevel(level)
    ten8t_logger.propagate = propagate


def ten8t_reset_logging():
    """
    Reset the logger to its initial state by removing all handlers and reinstating a NullHandler.

    NOTE: This function is intended for testing purposes ONLY.
          Using this in production code may disrupt logging configurations.
    """
    global ten8t_logger
    # Remove all existing handlers
    for handler in list(ten8t_logger.handlers):  # Make a copy to avoid modifying the list during iteration
        ten8t_logger.removeHandler(handler)

    # Add a NullHandler to restore the default silent state
    ten8t_logger.addHandler(logging.NullHandler())

    # Reset the logger's level and propagation settings
    ten8t_logger.setLevel(logging.NOTSET)
    ten8t_logger.propagate = False



"""
Allow the usage of an JSON file as an RC file.  
"""
import json
import pathlib

from .ten8t_exception import Ten8tException
from .ten8t_rc import Ten8tRC


class Ten8tJsonRC(Ten8tRC):
    """
    Loads configurations from JSON files. Extends Ten8tRC.
    """

    def __init__(self, cfg: str, section: str):
        section_data = self._load_config(cfg, section)

        self.expand_attributes(section_data)

    def _load_config(self, cfg: str, section: str) -> dict:
        """Loads and returns the requested section from a JSON file."""
        cfg_file = pathlib.Path(cfg)
        try:
            with cfg_file.open("rt", encoding="utf8") as j:
                config_data = json.load(j)
        except (FileNotFoundError, json.JSONDecodeError, AttributeError, PermissionError) as error:
            raise Ten8tException(f"JSON config {cfg} error: {error}") from error

        return config_data.get(section, {})



"""
Helper classes to support making environment variables with mutable values be less prone
to being changed by mistake.  We are under no presumption that we can stop people from
using a dynamic language.  These classes make the best effort to  prevent the user from
making edits to environment data that should be constant for the life of a rule
checking run.

THERE IS NO ASSURANCE THAT THIS WILL WORK IN ALL CASES. DON'T WRITE TO THE ENV VARIABLES!

"""

from .ten8t_exception import Ten8tException


class Ten8tEnvList(list):
    """
    Class representing a mutation-inhibited list. Mutational operations raise
    ten8t_exception.Ten8tException.

    Python being dynamic means forceful mutations can succeed. This class serves
    to prevent accidental changes by raising exceptions for mutating methods.

    Ideally, a copy is best to avoid mutation. But for large data sets, it's
    resource-demanding. ImmutableList protects large sets from unintended changes.
    """

    def __init__(self, *args):
        # super(Ten8tEnvList, self).__init__(*args)
        super().__init__(*args)

    def __setitem__(self, index, value):
        raise Ten8tException("Environment list does not support item assignment")

    def __delitem__(self, index):
        raise Ten8tException("Environment list doesn't support item deletion")

    def append(self, value):
        raise Ten8tException("Environment list is immutable, append is not supported")

    def extend(self, value):
        raise Ten8tException("Environment list is immutable, extend is not supported")

    def insert(self, index, value):
        raise Ten8tException("Environment list is immutable, insert is not supported")

    def remove(self, value):
        raise Ten8tException("Environment list is immutable, remove is not supported")

    def pop(self, index=-1):
        raise Ten8tException("Environment list is immutable, pop is not supported")

    def clear(self):
        raise Ten8tException("Environment list is immutable, clear is not supported")

    def sort(self, *args, **kwargs):
        raise Ten8tException("Environment list is immutable, sort is not supported")

    def reverse(self):
        raise Ten8tException("Environment list is immutable, reverse is not supported")


class Ten8tEnvDict(dict):
    """
    A class symbolizing a mutation-prohibited dictionary. Mutational operations raise
    a Ten8tException.

    Analogous to ImmutableList, Python's dynamic nature may allow forced mutations. This
    class prevents unintentional modifications to a dict object.
    """

    def __init__(self, *args, **kwargs):
        # super(Ten8tEnvDict, self).__init__(*args, **kwargs)
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        raise Ten8tException("Environment dict does not support item assignment")

    def __delitem__(self, key):
        raise Ten8tException("Environment dict doesn't support item deletion")

    def pop(self, k, d=None):
        raise Ten8tException("Environment dict is immutable, pop is not supported")

    def popitem(self):
        raise Ten8tException("Environment dict is immutable, popitem is not supported")

    def clear(self):
        raise Ten8tException("Environment dict is immutable, clear is not supported")

    def update(self, other=(), **kwargs):
        raise Ten8tException("Environment dict is immutable, update is not supported")

    def setdefault(self, key, default=None):
        raise Ten8tException("Environment dict is immutable, setdefault is not supported")


class Ten8tEnvSet(frozenset):
    """ Support immutable sets using frozenset """



"""
Public API for the Ten8t project.
"""
from .rule_files import rule_large_files  # noqa: F401
from .rule_files import rule_max_files  # noqa: F401
from .rule_files import rule_path_exists  # noqa: F401
from .rule_files import rule_paths_exist  # noqa: F401
from .rule_files import rule_stale_files  # noqa: F401
from .ten8t_attribute import attributes  # noqa: F401
from .ten8t_attribute import get_attribute  # noqa: F401
# from .ten8t_attribute import _convert_to_minutes # noqa:F401
from .ten8t_checker import Ten8tChecker  # noqa: F401
from .ten8t_checker import Ten8tDebugProgress  # noqa; F401
from .ten8t_checker import Ten8tNoProgress  # noqa; F401
from .ten8t_checker import Ten8tProgress  # noqa: F401
from .ten8t_checker import exclude_levels  # noqa: F401
from .ten8t_checker import exclude_phases  # noqa: F401
from .ten8t_checker import exclude_ruids  # noqa: F401
from .ten8t_checker import exclude_tags  # noqa: F401
from .ten8t_checker import keep_levels  # noqa: F401
from .ten8t_checker import keep_phases  # noqa: F401
from .ten8t_checker import keep_ruids  # noqa: F401
from .ten8t_checker import keep_tags  # noqa: F401
from .ten8t_exception import Ten8tException  # noqa: F401
from .ten8t_format import BM
from .ten8t_format import Ten8tBasicHTMLRenderer
from .ten8t_format import Ten8tBasicMarkdown
from .ten8t_format import Ten8tBasicRichRenderer
from .ten8t_format import Ten8tBasicStreamlitRenderer
from .ten8t_format import Ten8tMarkup
from .ten8t_format import Ten8tRenderText
# from .ten8t_exception import Ten8tTypeError  # noqa: F401
# from .ten8t_exception import Ten8tValueError  # noqa: F401
from .ten8t_function import Ten8tFunction  # noqa: F401
from .ten8t_immutable import Ten8tEnvDict  # noqa: F401
from .ten8t_immutable import Ten8tEnvList  # noqa: F401
from .ten8t_immutable import Ten8tEnvSet  # noqa: F401
from .ten8t_jsonrc import Ten8tJsonRC  # noqa: F401
from .ten8t_logging import ten8t_logger  # noqa: F401
from .ten8t_logging import ten8t_reset_logging  # noqa: F401
from .ten8t_logging import ten8t_setup_logging  # noqa: F401
from .ten8t_module import Ten8tModule  # noqa: F401
from .ten8t_package import Ten8tPackage  # noqa: F401
from .ten8t_rc import Ten8tRC  # noqa: F401
from .ten8t_rc_factory import ten8t_rc_factory  # noqa:F401
from .ten8t_result import TR  # noqa: F401
from .ten8t_result import Ten8tResult  # noqa: F401
from .ten8t_result import Ten8tYield  # noqa: F401
from .ten8t_result import overview  # noqa: F401
from .ten8t_ruid import empty_ruids  # noqa: F401
from .ten8t_ruid import module_ruids  # noqa: F401
from .ten8t_ruid import package_ruids  # noqa: F401
from .ten8t_ruid import ruid_issues  # noqa: F401
from .ten8t_ruid import valid_ruids  # noqa: F401
from .ten8t_score import ScoreBinaryFail  # noqa: F401
from .ten8t_score import ScoreBinaryPass  # noqa: F401
from .ten8t_score import ScoreByFunctionBinary  # noqa: F401
from .ten8t_score import ScoreByFunctionMean  # noqa: F401
from .ten8t_score import ScoreByResult  # noqa: F401
from .ten8t_score import ScoreStrategy  # noqa: F401
from .ten8t_thread import Ten8tThread  # noqa: F401
from .ten8t_tomlrc import Ten8tTomlRC  # noqa: F401
from .ten8t_util import any_to_int_list  # noqa: F401
from .ten8t_util import any_to_str_list  # noqa: F401
from .ten8t_util import str_to_bool  # noqa: F401

try:
    import narwhals as nw
    from .rule_ndf import rule_validate_ndf_schema  # noqa: F401
    from .rule_ndf import rule_validate_ndf_values_by_col  # noqa: F401
    from .rule_ndf import rule_ndf_columns_check  # noqa: F401
    from .rule_ndf import extended_bool  # noqa: F401
except ImportError:
    pass

# webapi using requests
try:
    import requests
    from .rule_webapi import rule_url_200  # noqa: F401
    from .rule_webapi import rule_web_api  # noqa: F401
except ImportError:
    pass

# ping rules
try:
    import ping3
    from .rule_ping import rule_ping_check  # noqa: F401
except ImportError:
    pass

# xlsx rules
try:
    import openpyxl
    from .rule_xlsx import rule_xlsx_a1_pass_fail
    from .rule_xlsx import rule_xlsx_df_pass_fail
except ImportError:
    pass

# pdf rules
try:
    import camelot  # type: ignore
    import pandas as pd  # pylint: disable=ungrouped-imports
    from .rule_pdf import extract_tables_from_pdf  # noqa: F401
    from .rule_pdf import rule_from_pdf_rule_ids  # noqa: F401
except ImportError:
    pass

# sql alchemy support
try:
    import sqlalchemy
    from .rule_sqlachemy import rule_sql_table_col_name_schema
    from .rule_sqlachemy import rule_sql_table_schema
except ImportError:
    pass



"""
This module is used to check for RUID files in the system.
"""

from collections import Counter

from .ten8t_module import Ten8tModule
from .ten8t_package import Ten8tPackage


def module_ruids(module: Ten8tModule) -> list[str]:
    """get a list of all the RUIDS in a module"""
    return [function.ruid for function in module.check_functions]


def package_ruids(package: Ten8tPackage) -> list[str]:
    """get a list of all the RUIDS in a package"""
    ruids = []
    for module in package.modules:
        ruids.extend(module_ruids(module))
    return ruids


def empty_ruids(ruids: list[str]) -> bool:
    """Check if all ruids are empty"""
    return all((not ruid for ruid in ruids))


def valid_ruids(ruids: list[str]) -> bool:
    """Check if all ruids are valid"""
    return len(ruids) == len(set(ruids)) and "" not in ruids


def ruid_issues(ruids: list[str]) -> str:
    """
    Provide a message about the issues with the RUIDs in the package.

    An issue is defined as either duplicate RUIDs (excluding blank entries)
    or the presence of blank RUIDs. The returned message will list any
    duplicate RUIDs and indicate whether blank RUIDs are present.

    If no issues are found, the returned message will indicate this.
    """
    ruid_counts = Counter(ruids)
    duplicates = [
        ruid for ruid, count in ruid_counts.items() if count > 1 and ruid != ""
    ]
    blanks = "" in ruids

    issues = []
    if duplicates:
        issues.append(f"Duplicates: {', '.join(duplicates)}.")
    if blanks:
        if empty_ruids(ruids):
            issues.append("RUIDS are not used.")
        else:
            issues.append("Blank RUIDs are present.")

    msg = " ".join(issues) if issues else "No issues found."
    return msg



"""
This is the sad place for lonely functions that don't have a place
"""


def str_to_bool(s: str, default=None) -> bool:
    """ Convert a string value to a boolean."""
    s = s.strip().lower()  # Remove spaces at the beginning/end and convert to lower case

    if s in ('pass', 'true', 'yes', '1', 't', 'y', 'on'):
        return True
    if s in ('fail', 'false', 'no', '0', 'f', 'n', 'off'):
        return False

    if default is not None:
        return default

    raise ValueError(f'Cannot convert {s} to a boolean.')


def any_to_str_list(param: str | list | None, sep=' ') -> list[str]:
    """
    Convert a string to a list of strings or if a list is given make sure it is all strings.
    Args:
        param: list of strings or string to convert to list of strings
        sep: separator character.

    Returns:

    """
    if param is None:
        return []
    if isinstance(param, str):
        param = param.strip()
        if param == '':
            return []
        else:
            return param.split(sep)
    if isinstance(param, list):
        if all(isinstance(item, str) for item in param):
            return param
    raise ValueError('Invalid parameter type, expected all strings.')


def any_to_int_list(param: str | list[int] | None, sep=' ') -> list[int]:
    """
    Convert a string to a list of integers or if a list is given make sure it is all integers.
    Args:
        param: list of integers or string to convert to list of integers
        sep: separator character.

    Returns:
        list of integers
    """
    if param is None:
        return []
    if isinstance(param, str):
        param = param.strip()
        cleaned_param = param.split(sep)
        try:
            return [int(x) for x in cleaned_param]
        except ValueError as exc:
            raise ValueError(
                'Invalid parameter value, expected numeric string values that can be converted to integers.') from exc
    if isinstance(param, list):
        return [int(x) for x in param]

    raise ValueError(f'Invalid parameter type in {param}, expected all integers.')



import copy
import threading
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed

from .ten8t_checker import Ten8tChecker
from .ten8t_result import TR, Ten8tResult


class Ten8tThread:
    """
    A class to manage and execute Ten8tChecker tasks in parallel using threading.

    This class groups functions collected by a Ten8tChecker instance based on their `thread_id`
    attribute, executes each group in parallel using `ThreadPoolExecutor`, and aggregates the results.

    Example:
        checker = Ten8tChecker(...)
        t_thread = Ten8tThread(checker)
        results = t_thread.run_all(max_workers=5)
    """

    def __init__(self, checker: Ten8tChecker):
        """
        Initialize the Ten8tThread instance with a specific Ten8tChecker.

        Args:
            checker (Ten8tChecker): An instance of Ten8tChecker containing the collected functions
                                    that need to be executed.
        """
        self.checker = checker

        # Organize the collected functions into groups based on their `thread_id`.
        self.thread_groups = self.make_thread_groups()

        self.results: list[Ten8tResult] = []

    @property
    def expected_threads(self) -> int:
        """
        Calculate the number of threads expected based on the number of unique thread groups.

        (`thread_id` values) present in `_func_groups`.

        Returns:
            int: Number of thread groups (keys) in `_func_groups`.
        """
        return len(self.thread_groups)

    def __repr__(self) -> str:
        """
        Return a string representation of the Ten8tThread instance.

        Includes:
            - Class name.
            - Number of thread groups (expected threads).
            - Checker's class name for context.

        Returns:
            str: A string suitable for debugging, showing the internal state of the object.
        """
        return (
            f"<{self.__class__.__name__}(expected_threads={self.expected_threads}, "
            f"checker={self.checker.__class__.__name__})>"
        )

    def make_thread_groups(self) -> defaultdict:
        """
        Groups the functions collected by the Ten8tChecker based on their `thread_id` attribute.
        The result of this is a dictionary where keys are `thread_id` values, and values are lists
        of functions corresponding to those thread IDs.

        Returns:
            defaultdict: A dictionary where keys are `thread_id` values, and values are lists
                         functions corresponding to those thread IDs.
        """

        fg = defaultdict(list)
        for function_ in self.checker.collected:
            fg[function_.thread_id].append(function_)
        self.thread_groups = fg
        return fg

    def run_all(self, max_workers=5) -> list[Ten8tResult]:
        """
        Execute all groups of functions in threads, where each group is in
        a dictionary keyed by its `thread_id`.

        If there is only one group, all functions are executed using the 'normal'
        checker object without threading. Otherwise, it uses a thread pool to execute
        each group concurrently.

        Args:
            max_workers (int): The maximum number of worker threads to use for parallel execution.
                               Default is 5.

        Returns:
            list[Ten8tResult]: A list of `Ten8tResult` objects representing the results of executed
                               check functions.
        """
        # If only one group of functions exists, execute them sequentially without threading.

        if len(self.thread_groups) == 1:
            return self.checker.run_all()

        # List to hold checkers, each assigned a subset of functions to process.
        checkers = []
        for name, functions in self.thread_groups.items():
            # Create a shallow copy of the Ten8tChecker instance.  This is a bit of paranoia on my part
            # but not enough to do deep copies.  Perhaps there are use cases for this?
            checker = copy.copy(self.checker)

            # Assign the subset of functions corresponding to the current group.
            checker.collected = functions

            # Add this modified checker to the list of checkers to process.
            checkers.append(checker)

        # Helper function to execute a checker's `run_all` method.
        def runner(checker_: Ten8tChecker) -> list[Ten8tResult]:
            """
            Execute the `run_all` method of the given Ten8tChecker instance.

            Args:
                checker_ (Ten8tChecker): The checker instance to execute.

            Returns:
                list[Ten8tResult]: The list of results from executing the checker.
            """

            # Assign the thread name of the checker object to the thread.
            # TODO: This is a hack.  When we run from Ten8tThread we know
            #       all thread_ids are the same so we are safe grabbing the first one
            #       this will need to be refactored.  This is ONLY done to make
            #       log messages know the thread_id.
            if len(checker_.collected) > 0:
                threading.current_thread().name = checker_.collected[0].thread_id
            else:
                threading.current_thread().name = "Checker"

            return checker_.run_all()

        final_result: list[Ten8tResult] = []  # This will collect all results from the threads.

        # Use ThreadPoolExecutor to execute each checker in parallel.
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # Submit all checkers to the thread pool for parallel execution.
            futures = [executor.submit(runner, checker) for checker in checkers]

            # Collect results from each thread as they complete.
            for future in as_completed(futures):
                try:
                    # Combine results from the completed thread into `final_result`.
                    final_result.extend(future.result())
                except Exception as e:  # pragma: no cover
                    # I don't know how to trigger this exception I have this code here to remind me what could go wrong
                    # This should be impossible, less out of memory type stuff, since the code in the checker handles
                    # all exceptions... ideally this code all goes away and this runner is trivial.
                    final_result.append(TR(status=False, msg=f"Unexpected exception in ten8t_thread.run_all {e} "))

        # Return the aggregated results from all threads.
        self.results = final_result

        # This might not be needed, but it is useful since threaded code returns the results interleaved.
        self.results.sort(key=lambda result: result.thread_id)

        return self.results



"""
Classes to support formatting result output.

It is assumed the output from ten8t will be used in different contexts, command line, web, markdown etc.  
As such, this class generalizes the output of ten8t's messages.

Blink message may use Ten8tMarkup to format them.  In most cases the raw text is sufficient, but there
are instances were nicer output is desired or required.

In order to accomplish this, the Ten8tMarkup class allows a very simple formatting of the messages.  THis
markup may be rendered in various ways dependent upon the message formatting.  This code provides basic markdown
that looks similar to HTML.

You can just write the markdown in your messages, or you can use the Ten8tMarkup class to generate
the inline text. 

The default formatting is plain text, which effectively just removes all formatting.

"""

from abc import ABC, abstractmethod

# Supported HTML style tags
# Define your tags as constants
TAG_BOLD = 'b'
TAG_ITALIC = 'i'
TAG_UNDERLINE = 'u'
TAG_STRIKETHROUGH = 's'
TAG_DATA = 'data'
TAG_EXPECTED = 'expected'
TAG_ACTUAL = 'actual'
TAG_FAIL = 'fail'
TAG_PASS = 'pass'
TAG_SKIP = 'skip'
TAG_WARN = 'warn'
TAG_CODE = 'code'
TAG_RED = 'red'
TAG_BLUE = 'blue'
TAG_GREEN = 'green'
TAG_PURPLE = 'purple'
TAG_ORANGE = 'orange'
TAG_YELLOW = 'yellow'
TAG_BLACK = 'black'
TAG_WHITE = 'white'


class Ten8tMarkup:
    """
    Baseline formatter class to be used by ten8t rules. 
    
    The idea of ten8t markup is a way to tag all result message with formatting information that
    may be used to provide the end user with a richer formatting experience targeting multiple
    output environments.  Since I use rich, markdown and streamlit and a bit of HTML I needed it
    to work for those platforms.  
    
    Ten8t messages can have embedded formatting information that indicates to higher level
    code how the text could be formatted.  Rendering class can choose what to do with the tags
    including doing nothing.
    
    The markup looks like html, with open and close tags.  This was chosen to make the simple
    search/replace algorithm have easy matches.
    
    The idea is that you could want a way to write a red message, so you could do:
    
    "Hello <<red>>world!<<red>>"
    
    or perhaps in code 
    
    f"Hello {fmt.red('world!')}"
    
    to make your formatted output.  
    
    If no formatter was specified, a text formatter would just run through the code and strip out
    all the <<>> tags, leaving the plain text:
    
    Hello world!
    
    A rich formatter might replace those tags with:
    
    "Hello [red]world![/red]"
    
    Incorporating this mechanism into your code should be pretty easy.  All you need to do
    is markup your text with ANY subset of the markup and then choose a render class to
    render output to your target.  Initially you use the text render which does
    nothing but strip off all the tags, leaving you with plain text messages suitable for
    logging.
    
    NOTE: This markup is intended for small amounts of text.  Generally single lines of text output.
    It should be used to add some color or bold text to line based string output.  It supports many outputs
    it sort of only works well for the subset of features that every target supports.  If you target HTML
    then you should be able to do almost anything since it supports deep nesting, however if you target
    markdown you will run into issues if you try deeply nest (or even nest) some tags.
       
    """

    # If you don't like my delimiters pick your own.

    def __init__(self, open_delim: str = "<<@>>", close_delim: str = "<</@>>"):
        """
        Init only allows you to specify open and close delimiters.
        
        The assumption is that you want markup that looks like <<bold>> <</bold>> and this
        __init__ defaults to those delimiters.  
        
        If you don't like the <<>> then you can change it.
        
        """
        self.open_delim = open_delim
        self.close_delim = close_delim

        if open_delim == close_delim:
            raise ValueError("Open and close delimiters for markup should not be the same.")

    def open_tag(self, tag: str) -> str:
        """if tag is 'red' open tag is <<red>>"""
        odl = self.open_delim.replace("@", tag.strip())
        return odl

    def close_tag(self, tag: str) -> str:
        """If  tag is 'red' close tag is <</red>>"""
        cdl = self.close_delim.replace("@", tag.strip())
        return cdl

    def _tag(self, id_, msg):
        """Create a generic tag string like <<code>>x=1<</code>>"""
        return f'{self.open_tag(id_)}{msg}{self.close_tag(id_)}'

    def bold(self, msg):
        """Create bold tag function. """
        return self._tag('b', msg)

    def italic(self, msg):
        """Create italic tag function. """
        return self._tag('i', msg)

    def underline(self, msg):
        """Create underline tag function."""
        return self._tag('u', msg)

    def strikethrough(self, msg):
        """Create strikethrough tag function. """
        return self._tag('s', msg)

    def code(self, msg):
        """Create code tag function. """
        return self._tag('code', msg)

    def data(self, msg):
        """Create data tag function. """
        return self._tag('data', msg)

    def expected(self, msg):
        """Create expected tag function. """
        return self._tag('expected', msg)

    def actual(self, msg):
        """Create actual tag function. """
        return self._tag('actual', msg)

    def fail(self, msg):
        """Create fail tag function. """
        return self._tag('fail', msg)

    def pass_(self, msg):
        """Create pass tag function. """
        return self._tag('pass', msg)

    def warn(self, msg):
        """Create warn tag function. """
        return self._tag('warn', msg)

    def skip(self, msg):
        """Create skip tag function. """
        return self._tag('skip', msg)

    def red(self, msg):
        """Create red tag function. """
        return self._tag('red', msg)

    def blue(self, msg):
        """Create blue tag function. """
        return self._tag('blue', msg)

    def green(self, msg):
        """Create green tag function. """
        return self._tag('green', msg)

    def yellow(self, msg):
        """Create yellow tag function. """
        return self._tag('yellow', msg)

    def orange(self, msg):
        """Create orange tag function. """
        return self._tag('orange', msg)

    def purple(self, msg):
        """Create purple tag function. """
        return self._tag('purple', msg)

    def black(self, msg):
        """Create black tag function. """
        return self._tag('black', msg)

    def white(self, msg):
        """Create white tag function. """
        return self._tag('white', msg)


# Create in instance of the markup class that can easily be used.  This instance
# is a shorthand that makes writing f-strings more compact and have access to a global
# markup formatter.  I have no thought that there will be multiple markups running at
# the same time, though I can image multiple renderers running at the same time, the
# obvious example being writing to a log file and a web interface or markdown file.
BM = Ten8tMarkup()


class Ten8tAbstractRender(ABC):
    """
    Base class for all ten8t renderers.  This has a list of all supported tags, the abstract
    render method and a concrete cleanup that removes all un-rendered tags.
    """

    # List of all known tags.  We need the list of all tags because code will need to run through all
    # tags and remove them if they aren't formatted.
    tags = [TAG_BOLD, TAG_ITALIC, TAG_UNDERLINE, TAG_STRIKETHROUGH, TAG_DATA, TAG_EXPECTED, TAG_ACTUAL, TAG_FAIL,
            TAG_PASS, TAG_CODE, TAG_RED, TAG_BLUE, TAG_GREEN, TAG_PURPLE, TAG_ORANGE, TAG_YELLOW, TAG_BLACK, TAG_WHITE,
            TAG_WARN, TAG_SKIP]

    @abstractmethod
    def render(self, msg):  # pragma: no cover
        """Base class render method"""

    def cleanup(self, msg):
        """
        It is optional for subclasses to replace all the render tags.  This method provides
        support to wipeout all un rendered tags.
        """
        fmt = Ten8tMarkup()

        # Find all the defined tags and blow them away.
        for tag in self.tags:
            msg = msg.replace(f'{fmt.open_tag(tag)}', '').replace(f'{fmt.close_tag(tag)}', '')
        return msg


class Ten8tRenderText(Ten8tAbstractRender):
    """
    This class strips all html formatting and is suitable for text output for things like CLI or API
    
    These messages are generally considered 'one-liners'
    """

    def render(self, msg):
        # Call the baseclass cleanup messages since this is a do-nothing class.
        return self.cleanup(msg)


class Ten8tBasicMarkdown(Ten8tRenderText):
    """Markdown render class"""

    def render(self, msg):
        """Basic markdown render method that converts tags to markdown"""
        fmt = Ten8tMarkup()
        replacements = {TAG_BOLD: '**',
                        TAG_ITALIC: '*',
                        TAG_STRIKETHROUGH: '~~',
                        TAG_CODE: '`',
                        TAG_PASS: '`',
                        TAG_FAIL: '`',
                        TAG_WARN: '`',
                        TAG_SKIP: '`',
                        TAG_EXPECTED: '`',
                        TAG_ACTUAL: '`'}
        # Phase 1 replace all substitutions
        for md, replacement in replacements.items():
            open_tag = fmt.open_tag(md)
            close_tag = fmt.close_tag(md)
            msg = msg.replace(open_tag, replacement).replace(close_tag, replacement)

        # Phase 2 replace all unused tags.
        return self.cleanup(msg)


class Ten8tBasicRichRenderer(Ten8tRenderText):
    """Rich render class"""

    def render(self, msg):
        """Basic markdown render method that converts tags to rich formatted code"""
        fmt = Ten8tMarkup()
        replacements = {TAG_BOLD: ('[bold]', '[/bold]'),
                        TAG_ITALIC: ('[italic]', '[/italic]'),
                        TAG_UNDERLINE: ('[u]', '[/u]'),
                        TAG_STRIKETHROUGH: ('[strike]', '[/strike]'),
                        TAG_PASS: ('[green]', '[/green]'),
                        TAG_FAIL: ('[red]', '[/red]'),
                        TAG_WARN: ('[orange]', '[/orange]'),
                        TAG_SKIP: ('[purple]', '[/purple]'),
                        TAG_EXPECTED: ('[green]', '[/green]'),
                        TAG_ACTUAL: ('[green]', '[/green]'),
                        TAG_RED: ('[red]', '[/red]'),
                        TAG_GREEN: ('[green]', '[/green]'),
                        TAG_BLUE: ('[blue]', '[/blue]'),
                        TAG_YELLOW: ('[yellow]', '[/yellow]'),
                        TAG_ORANGE: ('[orange]', '[/orange]'),
                        TAG_PURPLE: ('[purple]', '[/purple]'),
                        TAG_BLACK: ('[black]', '[/black]'),
                        TAG_WHITE: ('[white]', '[/white]'),
                        }

        # Phase 1 replace all substitutions
        for md, replacement in replacements.items():
            msg = msg.replace(f"{fmt.open_tag(md)}", replacement[0]).replace(f"{fmt.close_tag(md)}", replacement[1])

        # Phase 2 replace all unused tags.
        return self.cleanup(msg)


class Ten8tBasicStreamlitRenderer(Ten8tRenderText):
    """Streamlit renderer class."""

    def render(self, msg):
        """Basic markdown render method that converts tags to streamlit format strings"""

        fmt = Ten8tMarkup()
        replacements = {
            TAG_BOLD: ('**', '**'),
            TAG_ITALIC: ('*', '*'),
            TAG_CODE: ('`', '`'),
            TAG_PASS: (':green[', ']'),
            TAG_FAIL: (':red[', ']'),
            TAG_WARN: (':orange[', ']'),
            TAG_SKIP: (':purple[', ']'),
            TAG_EXPECTED: (':green[', ']'),
            TAG_ACTUAL: (':green[', ']'),
            TAG_RED: (':red[', ']'),
            TAG_GREEN: (':green[', ']'),
            TAG_BLUE: (':blue[', ']'),
            TAG_YELLOW: (':yellow[', ']'),
            TAG_ORANGE: (':orange[', ']'),
            TAG_PURPLE: (':purple[', ']'),
            TAG_BLACK: (':black[', ']'),
            TAG_WHITE: (':white[', ']'),
        }

        # Phase 1 replace all substitutions
        for md, replacement in replacements.items():
            open_tag = fmt.open_tag(md)
            close_tag = fmt.close_tag(md)

            msg = msg.replace(open_tag, replacement[0]).replace(close_tag, replacement[1])

        # Phase 2 replace all unused tags.
        msg = self.cleanup(msg)

        return msg


class Ten8tBasicHTMLRenderer(Ten8tRenderText):
    """HTML renderer"""

    def render(self, msg):
        """Basic markdown render method that converts tags to html"""

        fmt = Ten8tMarkup()
        replacements = {
            TAG_BOLD: ('<b>', '</b>'),
            TAG_ITALIC: ('<i>', '</i>'),
            TAG_UNDERLINE: ('<u>', '</u>'),
            TAG_STRIKETHROUGH: ('<s>', '</s>'),
            TAG_CODE: ('<code>', '</code>'),
            TAG_PASS: ('<span style="color:green">', '</span>'),
            TAG_FAIL: ('<span style="color:red">', '</span>'),
            TAG_WARN: ('<span style="color:orange">', '</span>'),
            TAG_SKIP: ('<span style="color:purple">', '</span>'),
            TAG_EXPECTED: ('<span style="color:green">', '</span>'),
            TAG_ACTUAL: ('<span style="color:red">', '</span>'),
            TAG_RED: ('<span style="color:red">', '</span>'),
            TAG_GREEN: ('<span style="color:green">', '</span>'),
            TAG_BLUE: ('<span style="color:blue">', '</span>'),
            TAG_YELLOW: ('<span style="color:yellow">', '</span>'),
            TAG_ORANGE: ('<span style="color:orange">', '</span>'),
            TAG_PURPLE: ('<span style="color:purple">', '</span>'),
            TAG_BLACK: ('<span style="color:black">', '</span>'),
            TAG_WHITE: ('<span style="color:white">', '</span>'),
        }

        # Phase 1 replace all substitutions
        for md, replacement in replacements.items():
            open_tag = fmt.open_tag(md)
            close_tag = fmt.close_tag(md)

            msg = msg.replace(open_tag, replacement[0]).replace(close_tag, replacement[1])

        # Phase 2 replace all unused tags.
        msg = self.cleanup(msg)

        return msg



"""
Set of baseline rules that uses the pyfilesystem module to OS-agnostic checks on things
about the file system, file existing, age etc.
"""

from typing import Generator, Sequence

from sqlalchemy import Engine, MetaData, Table
from sqlalchemy.sql.type_api import TypeEngine

from .ten8t_format import BM
from .ten8t_result import TR


def rule_sql_table_col_name_schema(engine: Engine,
                                   table: str,
                                   expected_columns: Sequence[str],
                                   extra_columns_ok: bool = True) -> Generator[TR, None, None]:
    """
    Take a connection, a table, and column names and verify that the table has those columns.
    Note that this does NOT verify the data types of the columns.
    Args:
        engine: SQLAlchemy engine object
        table: Name of the table
        expected_columns: List of expected column names
        extra_columns_ok: Boolean if extra columns are OK.  False=Exact match required

    Returns:
        A generator yielding assertion results for each column
    """

    if not table:
        yield TR(status=False, msg="Table name cannot be blank.")
        return

    if not expected_columns:
        yield TR(status=False, msg="Column list cannot be empty.")
        return

    count = 0

    for expected_column in expected_columns:
        if not expected_column.strip():
            yield TR(status=False, msg="Column names cannot be empty.")
            count += 1

    if count:
        return

    # Get the table's metadata
    metadata = MetaData()
    table_obj = Table(table, metadata, autoload_with=engine)
    actual_columns = set(table_obj.columns.keys())

    # Verify expected columns exist
    for column in expected_columns:
        if column in actual_columns:
            yield TR(status=True, msg=f"Column {BM.code(column)} is present in table {BM.code(table)}")
        else:
            yield TR(status=False, msg=f"Column {BM.code(column)} is {BM.fail('MISSING')} in table {BM.code(table)}")

    # If extra columns existing in the database is OK then don't check
    if not extra_columns_ok:
        extra_columns = actual_columns - set(expected_columns)

        # This is very subtle.  We want the ordering of actual columns here  rather than iterating
        # over the set (that varies with python version).   For small sets this isn't to terrible
        for column in actual_columns:
            if column in extra_columns:
                yield TR(status=False, msg=f"Column {BM.code(column)} is UNEXPECTED in table {BM.code(table)}")


def rule_sql_table_schema(engine: Engine,
                          table: str,
                          expected_columns: list[tuple[str, TypeEngine]],
                          extra_columns_ok: bool = True) -> Generator[TR, None, None]:
    """
    Take a connection, a table, and column names and verify that the table has those columns
    AND the correct types.
    Args:
        engine: SQLAlchemy engine object
        table: Name of the table
        expected_columns: List of expected column name,type pairs
        extra_columns_ok: (default=True) Ignore additional columns in the table.

    Returns:
        A generator yielding assertion results for each column
    """
    if not table:
        yield TR(status=False, msg="Table name cannot be blank.")
        return

    if not expected_columns:
        yield TR(status=False, msg="Column list cannot be empty.")
        return

    # Get the table's metadata
    metadata = MetaData()
    table_obj = Table(table, metadata, autoload_with=engine)
    actual_columns = {column.name: column.type for column in table_obj.columns}

    # Verify expected columns exist and have correct types
    for expected_column, expected_type in expected_columns:
        actual_type = actual_columns.get(expected_column)

        # Check if column exists
        if actual_type:
            # Remove any qualifiers from the SQLAlchemy type to get base type
            unqualified_actual_type = actual_type.__class__.__name__
            unqualified_expected_type = expected_type.__class__.__name__

            # Check if types match
            if unqualified_actual_type == unqualified_expected_type:
                # pylint: disable=line-too-long
                yield TR(status=True,
                         msg=f"Column {BM.expected(expected_column)} of type {BM.expected(expected_type)} "
                             f"is correctly present in table {BM.code(table)}")
            else:
                # pylint: disable=line-too-long
                yield TR(status=False,
                         msg=f"Column {BM.expected(expected_column)}  has incorrect type. Expected: "
                             f"{BM.expected(unqualified_expected_type)} , got: {BM.actual(unqualified_actual_type)}")
        else:
            yield TR(status=False, msg=f"Missing column in table {table}: {BM.expected(expected_column)} ")

    # If extra columns existing in the database is OK then don't check
    if not extra_columns_ok:
        extra_columns = set(actual_columns.keys()) - set(column for column, _ in expected_columns)
        for column in extra_columns:
            yield TR(status=False, msg=f"Unexpected column in table {BM.code(table)}: {BM.code(column)}")



"""
This class encapsulates the discovered rule functions found in the system.  A great
deal of metadata is stored in the function and extracted from information about the function
its signature, its generator status etc.  This information is used so users do not need to
configure functions in multiple places.  Design elements from fastapi and pytest are obvious.
"""
import inspect
import re
import time
import traceback
from typing import Any, Generator

from .ten8t_attribute import get_attribute
from .ten8t_exception import Ten8tException
from .ten8t_logging import ten8t_logger
from .ten8t_result import Ten8tResult


def result_hook_fix_blank_msg(sfunc: "Ten8tFunction",
                              result: Ten8tResult) -> Ten8tResult:
    """Fix the message of a result if it is blank.

    This is an example of a result hook used by ten8t to write
    a useful message if the user did not provide one.

    Args:
        sfunc (Ten8tFunction): The function to fix       
        result (Ten8tResult): The result to rewrite.

    Returns:
        Ten8tResult: The result with the message fixed.
    """
    # If the result has no message, then create a default one either
    # from the doc string or from the function name/module/package.
    if not result.msg:
        if sfunc.doc and sfunc.doc.strip().count("\n") == 0 and sfunc.doc.strip():
            result.msg = f"{result.func_name}: {sfunc.doc.strip()}"
        else:
            # Use a dictionary to store the optional parts of the message
            # This makes it easier to add or remove parts in the future
            msg_parts = {
                "tag": sfunc.tag,
                "level": sfunc.level,
                "phase": sfunc.phase,
                "module": sfunc.module,
            }
            # Only include the parts that have a value
            msg_str = " ".join(
                f"{key}={value}" for key, value in msg_parts.items() if value
            )
            result.msg = f"Ran {sfunc.function_name}.{result.count:03d} {msg_str}"
    return result


ATTRIBUTES = ("tag", "level", "phase", "weight", "skip", "ruid", "skip_on_none",
              "fail_on_none", "ttl_minutes", "finish_on_fail")


class Ten8tFunction:
    """
        A class representing a function within the Ten8t framework's module.

        The class stores a function, its module and allows error-handled function calls.
        All functions returning values convert to generators yielding Ten8tResult objects.
        This ensures a consistent return type, eliminating the need to worry about
        functions returning multiple or single results.

        Ideal use is with generated generators, which attach useful attributes
        to each function. These attributes are then processed into the Ten8tFunction object.

        Attributes:
        - module (ModuleType): Contains the function.
        - function (Callable): Function to be called.
        - parameters (inspect.Parameters): Function parameters.
        - allowed_exceptions (Tuple[Type[Exception]]): Exception types for try-catch clauses.

        Methods:
        - __str__(): Returns string representation of Ten8tFunction.
        - __call__(*args, **keywords): Calls the function and gathers result info.
        """

    def __init__(self, function_: Any,
                 module: str = '',
                 allowed_exceptions: tuple[type[BaseException], ...] = None,  # So mypy understands types
                 env: dict[Any, Any] = None,
                 pre_sr_hooks: Any = None,
                 post_sr_hooks: Any = None):
        self.env = env or {}
        self.module = module
        self.function = function_
        self.is_generator = inspect.isgeneratorfunction(function_)
        self.function_name = function_.__name__

        # Using inspect gets the docstring without the python indent.
        self.doc = inspect.getdoc(function_) or ""

        # Store parameter names so they can be filled from environment
        self.parameters = inspect.signature(function_).parameters
        self.result_hooks = [result_hook_fix_blank_msg]

        # Allow user to control rewriting the result hooks
        if isinstance(pre_sr_hooks, list):
            self.result_hooks = pre_sr_hooks + self.result_hooks
        elif pre_sr_hooks is None:
            pass
        else:
            raise Ten8tException("pre_sr_hooks must be a list")

        if isinstance(post_sr_hooks, list):
            self.result_hooks += post_sr_hooks
        elif post_sr_hooks is None:
            pass
        else:
            raise Ten8tException("post_sr_hooks must be a list")

        # This should be a class rather than having to repeat yourself.
        self.tag: str = get_attribute(function_, "tag")
        self.level: int = get_attribute(function_, "level")
        self.phase: str = get_attribute(function_, "phase")
        self.weight: float = get_attribute(function_, "weight")
        self.skip: bool = get_attribute(function_, "skip")
        self.ruid: str = get_attribute(function_, "ruid")
        self.skip_on_none: bool = get_attribute(function_, "skip_on_none")
        self.fail_on_none: bool = get_attribute(function_, "fail_on_none")
        self.ttl_minutes: float = get_attribute(function_, "ttl_minutes")
        self.finish_on_fail: bool = get_attribute(function_, "finish_on_fail")
        self.index = get_attribute(function_, "index")
        self.thread_id = get_attribute(function_, "thread_id")

        # Support Time To Live using the return value of time.time.  Resolution of this
        # is on the order of 10e-6 depending on OS.  In my case this is WAY more than I
        # need, and I'm assuming you aren't building a trading system with this, so you don't
        # care about microseconds.
        self.last_ttl_start: float = 0.0  # this will be compared to time.time() for ttl caching
        self.last_results: list[Ten8tResult] = []

        if self.weight in [True, False, None]:
            raise Ten8tException("Boolean and none types are not allowed for weights.")

        if self.weight <= 0.0:
            raise Ten8tException("Weight must be greater than 0.0")

        # This allows the library user to control how lenient the underlying code is
        # with exceptions.  This is a pain point in the implementation since we don't
        # really know what exceptions should be caught and indicated in the results
        # and which ones should cause the system to exit since this is a library, not
        # final application code.
        self.allowed_exceptions = allowed_exceptions or (Exception,)

    def __str__(self):
        return f"Ten8tFunction({self.function_name=})"

    def _get_parameter_values(self):
        args = []
        for param in self.parameters.values():
            if param.name in self.env:
                args.append(self.env[param.name])
            elif param.default != inspect.Parameter.empty:
                args.append(param.default)

        return args

    def _cache_result(self, result):
        """Simple caching saves results if ttl_minutes is no 0"""
        if self.ttl_minutes:
            self.last_results.append(result)

    def __call__(self, *args, **kwargs) -> Generator[Ten8tResult, None, None]:
        """Call the user provided function and collect information about the result.

        This is the heart of Ten8t.  Each of these functions checks something
        in the system using the provided function and feedback.  Each function is
        a generator (or just a function that pretends to be a generator). This code
        manages the details that we'd prefer to handle in the core of the system
        rather than inside the check functions.

        Raises:
            Ten8tException: Exceptions are remapped to Ten8tExceptions for easier handling

        Returns:
            Ten8tResult:

        Yields:
            Iterator[Ten8tResult]:
        """
        # Call the stored function and collect information about the result
        start_time: float = time.time()

        # Function returns a generator that needs to be iterated over
        args = self._get_parameter_values()

        # If any arguments are None that is a bad thing.  That means that
        # a file could not be opened or other data is not available. If
        # Functions are not allowed to update the environment this only
        # needs to run once, rather than on every function call
        for count, arg in enumerate([arg for arg in args if arg is None], start=1):

            # Make a nice message if there is a ruid for this rule
            ruid_msg = f'|{self.ruid}' if self.ruid else ''

            if self.fail_on_none:
                yield Ten8tResult(status=False,
                                  msg=f"Failed due to None arg. {count} in func='{self.function_name}'{ruid_msg}",
                                  fail_on_none=True)
                return
            if self.skip_on_none:
                yield Ten8tResult(status=None, skipped=True,
                                  msg=f"Skipped due to None arg. {count} in func='{self.function_name}{ruid_msg}'",
                                  skip_on_none=True)
                return

        # It is possible for an exception to occur before the generator is created.
        # so we need a value to be set for count.
        count = 1

        # If we need values from the result cache, then we can just yield them back
        if self.ttl_minutes * 60 + self.last_ttl_start > time.time():
            yield from self.last_results
            return

        try:
            self.last_results = []
            self.last_ttl_start = start_time

            # This allows for returning a single result using return or
            # multiple results returning a list of results.
            if not self.is_generator:
                # If the function is not a generator, then just call it
                results = self.function(*args)
                end_time = time.time()
                if isinstance(results, Ten8tResult):
                    results = [results]

                # TODO: I could not make a decorator work for this, so I just put it here.
                #       Ideally the attribute decorator could see a non generator function
                #       and wrap at creation rather than having this crap here.
                elif isinstance(results, bool):
                    results = [Ten8tResult(status=results)]
                if not isinstance(results[0], Ten8tResult):
                    raise Ten8tException(f"Invalid return from ten8t function {self.function_name}")
                for count, r in enumerate(results, start=1):
                    # TODO: Time is wrong here, we should estimate each part taking
                    #       1/count of the total time
                    r = self.load_result(r, start_time, end_time, count=1)
                    yield r

                    self._cache_result(r)

            else:
                # Functions can return multiple results, track them with a count attribute.
                for count, result in enumerate(self.function(*args, **kwargs), start=1):
                    end_time = time.time()

                    if isinstance(result, bool):
                        result = Ten8tResult(status=result)
                    elif isinstance(result, list):
                        raise Ten8tException(
                            "Function yielded a list rather than a Ten8tResult or boolean"
                        )

                    result = self.load_result(result, start_time, end_time, count)

                    yield result

                    self._cache_result(result)

                    start_time = time.time()

        except self.allowed_exceptions as e:
            # Generically handle exceptions here so we can keep running.
            result = Ten8tResult(status=False)
            result = self.load_result(result, 0, 0, count)
            result.except_ = e
            result.traceback = traceback.format_exc()
            mod_msg = "" if not self.module else f"{self.module}"
            result.msg = f"Exception '{e}' occurred while running {mod_msg}.{self.function.__name__}"
            ten8t_logger.error(result.msg)
            yield result

    def _get_section(self, header="", text=None):
        """
        Extracts a section from the docstring based on the provided header.
        If no header is provided, it returns the text before the first header.
        If the header is provided, it returns the text under that header.
        If the header isn't found, it returns the text before the first header.

        Parameters:
        header (str): The header of the section to extract.
        text (str): IInternally this is never used, but is useful for testing.

        Returns:
        str: The text of the requested section.
        """

        text = text or self.doc

        # Split the docstring into sections
        sections = re.split(r"(^\w+:)", text, flags=re.MULTILINE)

        # Just return the first line of the header
        if not header:
            return sections[0].strip().split("\n", 1)[0].strip()

        # Ensure the header ends with ':'
        header = header.strip() + ":" if not header.endswith(":") else header

        # Try to find the header and return the next section
        for i, section in enumerate(sections):
            if section.strip() == header:
                return sections[i + 1].strip()

        # If the header wasn't found, return the text before the first header
        return ""

    def load_result(self, result: Ten8tResult, start_time, end_time, count=1):
        """
        Provide a bunch of metadata about the function call, mostly hoisting
        parameters from the functon to the result.

        A design decision was made to make the result data flat since there are more than
        1 possible hierarchy.  Tall-skinny data that can be transformed into wide or
        hierarchical.
        """
        # Use getattr to avoid repeating the same pattern of checking if self.module exists
        result.pkg_name = getattr(self.module, "__package__", "")
        result.module_name = getattr(self.module, "__name__", "")

        # Assign the rest of the attributes directly
        result.ruid = self.ruid
        result.func_name = self.function_name
        result.doc = self.doc
        result.tag = self.tag
        result.level = self.level
        result.phase = self.phase
        result.runtime_sec = end_time - start_time
        result.ttl_minutes = self.ttl_minutes
        result.count = count
        result.thread_id = self.thread_id

        result.fail_on_none = self.fail_on_none
        result.skip_on_none = self.skip_on_none

        # Apply all (usually 1 or 0) hooks to the result
        for hook in self.result_hooks:
            if result is not None:
                result = hook(self, result)

        return result



"""
This module contains ten8t rules that are useful for checking the status of files
on the native file system.  This functions me be removed in a future release
and replaced by functions that use the pyfilesystem package.
"""
import pathlib
import time
from typing import Generator

from .ten8t_exception import Ten8tException
from .ten8t_format import BM
from .ten8t_result import TR, Ten8tYield


def rule_path_exists(path_: str) -> Generator[TR, None, None]:
    """Simple rule to check for a file path."""
    path_str = BM.code(path_)
    if pathlib.Path(path_).exists():
        yield TR(status=True, msg=f"The path {BM.code(path_str)} does exist.")
    else:
        yield TR(status=False, msg=f"The path  {BM.code(path_str)} does {BM.bold('NOT')} exist.")


def rule_paths_exist(paths: list[str] | str,
                     summary_only=False,
                     summary_name=None,
                     name="Path Check",
                     no_paths_pass_status=False) -> Generator[TR, None, None]:
    y = Ten8tYield(summary_only=summary_only, summary_name=summary_name)

    if isinstance(paths, str):
        paths = paths.split(" ")

    for path in paths:
        yield from y(rule_path_exists(path))

    if y.count == 0:
        yield from y(status=no_paths_pass_status,
                     msg=f"There were no paths to check in {BM.code(name)}.")

    if summary_only:
        yield from y.yield_summary()


def rule_stale_file(
        filepath: pathlib.Path,
        days: float = 0,
        hours: float = 0,
        minutes: float = 0,
        seconds: float = 0,
        current_time=None
) -> Generator[TR, None, None]:
    """Check a single file for being stale."""

    current_time = current_time or time.time()

    age_in_seconds = days * 86400.0 + hours * 3600.0 + minutes * 60.0 + seconds
    if age_in_seconds <= 0:
        raise Ten8tException(f"Age for stale file check {BM.code(age_in_seconds)} should be > 0")

    try:
        file_mod_time = filepath.stat().st_mtime
        file_age_in_seconds = current_time - file_mod_time

        if file_age_in_seconds > age_in_seconds:
            unit = "seconds"
            if days > 0:
                file_age = file_age_in_seconds / 86400.0
                unit = "days"
            elif hours > 0:
                file_age = file_age_in_seconds / 3600.0
                unit = "hours"
            elif minutes > 0:
                file_age = file_age_in_seconds / 60.0
                unit = "minutes"
            elif seconds > 0:
                file_age = file_age_in_seconds

            age_msg = f"age = {file_age:.2f} {unit} {age_in_seconds=}"
            yield TR(status=False, msg=f"Stale file {BM.code(filepath)} {BM.code(age_msg)}")
        else:
            yield TR(status=True, msg=f"Not stale file {BM.code(filepath)}")
    except (FileNotFoundError, PermissionError, IOError) as exc:
        yield TR(status=False,
                 msg="Exception occurred while checking for the path {SM.code(path_str)}",
                 except_=exc)


def rule_stale_files(
        folder: str | pathlib.Path,
        pattern: str | pathlib.Path,
        days: float = 0,
        hours: float = 0,
        minutes: float = 0,
        seconds: float = 0,
        no_files_pass_status: bool = True,
        summary_only=False,
        summary_name=None
) -> Generator[TR, None, None]:
    """
        Rule verifies no files older than a specified age. Each too-old file is reported.
        Age defined in days, hours, minutes, and seconds.

        No files found could be deemed pass or fail. This behavior can be set, with True as default.
    """
    y = Ten8tYield(summary_only=summary_only, summary_name=summary_name or "Rule_stale_files")

    current_time = time.time()
    for filepath in pathlib.Path(folder).rglob(str(pattern)):
        yield from y.results(rule_stale_file(filepath=filepath,
                                             days=days,
                                             hours=hours,
                                             minutes=minutes,
                                             seconds=seconds,
                                             current_time=current_time))

    if y.count == 0:
        yield from y(status=no_files_pass_status,
                     msg=f"No files were found in {BM.code(folder)} matching pattern {BM.code(folder)}")

    if summary_only:
        yield from y.yield_summary()


def rule_large_files(folder: str,
                     pattern: str,
                     max_size: float,
                     no_files_pass_status: bool = True,
                     summary_only=False,
                     summary_name=None):
    """
    Rule to verify that there are no files larger than a given size.
    Each file that is too big is reported.
    """

    y = Ten8tYield(summary_only=summary_only, summary_name=summary_name or "Rule_large_files")

    if max_size <= 0:
        raise Ten8tException(f"Size for large file check should be > 0 not {max_size=}")

    for filepath in pathlib.Path(folder).rglob(pattern):
        size_bytes = filepath.stat().st_size
        if size_bytes > max_size:
            yield from y(
                status=False,
                msg=f"Large file {BM.code(filepath)}, {BM.code(size_bytes)} bytes, exceeds limit of {BM.code(max_size)} bytes",
            )
    if y.count == 0:
        yield from y(status=no_files_pass_status,
                     msg=f"No files found matching {BM.code(pattern)} in {BM.code(folder)}.")

    if summary_only:
        yield from y.yield_summary()


def rule_max_files(folders: list,
                   max_files: list | int,
                   pattern: str = '*',
                   summary_only=False,
                   summary_name=None):
    """
    Rule to verify that the number of files in a list of folders does not
    exceed a given limit.
    """
    y = Ten8tYield(summary_only=summary_only, summary_name=summary_name or "Rule_max_files")

    if isinstance(folders, (str, pathlib.Path)):
        folders = [folders]
    if isinstance(max_files, int):
        max_files = [max_files] * len(folders)

    if len(folders) != len(max_files):
        raise Ten8tException(f"Number of folders and max_files {max_files} must be the same.")

    for folder, max_file in zip(folders, max_files):
        count = 0
        # don't materialize the list, just count
        for count, _ in enumerate(pathlib.Path(folder).rglob(pattern), start=1):
            pass

        if count <= max_file:
            yield from y(status=True,
                         msg=f"Folder {BM.code(folder)} contains less than or equal to {BM.code(max_file)} files.")
        else:
            yield from y(status=False,
                         msg=f"Folder {BM.code(folder)} contains greater than {BM.code(max_file)} files.")

    if summary_only:
        yield from y.yield_summary()



"""
Allow the usage of an INI file as an RC file.  
"""
import configparser

from .ten8t_exception import Ten8tException
from .ten8t_rc import Ten8tRC


class Ten8tIniRC(Ten8tRC):
    """
    Loads configurations from TOML files. Extends Ten8tRC.
    """

    def __init__(self, cfg: str, section: str):
        section_data = self._load_config(cfg, section)

        self.expand_attributes(section_data)

    def _load_config(self, cfg: str, section: str) -> dict:
        """Loads and returns the requested section from a TOML file."""
        try:
            config = configparser.ConfigParser()
            config.read(cfg, encoding="utf-8")
            if not section:
                raise Ten8tException("Section must be provided to read INI RC files.")
            d = {"tags": config.get(section, "tags"),
                 "ruids": config.get(section, "ruids"),
                 "phases": config.get(section, "phases")}

        except (FileNotFoundError, TypeError, configparser.Error, PermissionError) as error:
            raise Ten8tException(f"INI config file {cfg} error: {error}") from error

        return d



"""
Ten8tModule represents a module with a functions set that can be run. A module
typically symbolizes a file imported into the system. It does this by identifying
all functions starting with a certain prefix and adding them to a list managed
by ten8t.
"""

import importlib
import inspect
import pathlib
import sys
from collections import Counter

from .ten8t_exception import Ten8tException
from .ten8t_function import Ten8tFunction


class Ten8tModule:
    """
    A module is a collection of functions that is read from a file.  The check_ functions
    are used to verify rules, while the env_functions are used to set up any parameters
    that the rule functions might need.
    """

    def __init__(
            self,
            module_name: str,
            module_file: str,
            check_prefix="check_",
            env_prefix="env_",
            env_functions: list | None = None,
            auto_load=True,
    ) -> None:
        self.module_name: str = module_name
        self.check_functions: list[Ten8tFunction] = []
        self.env_functions: list = env_functions or []
        self.module = None
        self.module_file: str = module_file
        self.check_prefix: str = check_prefix
        self.env_prefix: str = env_prefix
        self.autothread: str = ""
        self.doc = ""
        if auto_load:
            self.load()

    def __str__(self):
        return (
            f"Ten8tModule({self.module_name=},{self.check_function_count=} functions)"
        )

    @property
    def check_function_count(self) -> int:
        """Return the check function count..."""
        return len(self.check_functions) if self.check_functions else 0

    def add_check_function(self, module, function):
        """Wrap the function in a ten8t function"""
        function = Ten8tFunction(function, module)
        self.check_functions.append(function)

    def add_env_function(self, func):
        """Add a discovered environment function to the list"""
        self.env_functions.append(func)

    @staticmethod
    def _add_sys_path(module_file: str | pathlib.Path) -> list[str]:
        """
        Add a module's directory to sys.path.
        If it is already there do nothing. 
        """

        # Construct a Path object from the provided file path and get its parent directory
        module_dir = pathlib.Path(module_file).parent.resolve()

        # Check if the module directory is already in sys.path
        if module_dir not in (pathlib.Path(path).resolve() for path in sys.path):
            sys.path.insert(0, str(module_dir))

        return sys.path

    # If not, add it to sys.path

    def load(self, module_name=None):
        """Load a module using importlib."""
        module_name = module_name or self.module_name
        self._add_sys_path(self.module_file)
        try:
            module = importlib.import_module(module_name)
            self.module = module
            # self.doc = module.__doc__
            self.doc = inspect.getdoc(module)
            self.load_special_functions(module)
            return True

        except ImportError as iex:
            raise Ten8tException(f"Can't load {module_name}:{iex.msg}") from iex

    def load_special_functions(self, module):
        """Look through all the functions in the module and load the check/env functions"""

        module = module or self.module

        # Dir givens you every element in the module object and allows us to look
        # for check functions that look a certain way
        check_func_count = 0
        for name in dir(module):
            if name.startswith("_"):
                continue
            obj = getattr(module, name)

            if not callable(obj):
                continue

            # Load environment functions
            if name.startswith(self.env_prefix):
                self.add_env_function(obj)

            # Load check functions.  We number them because we care about the 'file' order
            # The file order will let us sort/resort functions and access file order if we want it.
            if name.startswith(self.check_prefix):
                check_func_count += 1
                obj.index = check_func_count
                self.add_check_function(module, obj)

        # Strictly speaking this doesn't need to happen here, it could be checked later
        duplicate_ruids = [
            item
            for item, count in Counter(self.ruids()).items()
            if count > 1 and item != ""
        ]

        if duplicate_ruids:
            raise Ten8tException(
                f"Duplicate RUIDs found in module: {','.join(duplicate_ruids)}"
            )

    def ruids(self):
        """
        Return a list of all the RUIDs in the module.
        Note that this can have duplicates.  The list is
        sorted to facilitate comparison.

        RUID = rule identifier
        """
        return sorted(function.ruid for function in self.check_functions)



"""
Allow the usage of an XML file as an RC file.  Not sure why you want this.
"""
import pathlib
from xml.etree import ElementTree

from .ten8t_exception import Ten8tException
from .ten8t_rc import Ten8tRC


class Ten8tXMLRC(Ten8tRC):
    """
    Loads configurations from XML files. Extends Ten8tRC.

    This class is VERY simple at this time and will work only for the most
    trivial XML files.  I'm happy to not need XML much anymore.
    """

    def __init__(self, cfg: str, section: str):
        package_data = self._load_config(cfg, section)

        # Convert the +/-attributes into separate inclusion/exclusion lists.
        self.expand_attributes(package_data)

    def _load_config(self, cfg: str, section: str) -> dict:
        """Loads tags, phases and ruids from a given package in the XML file."""
        cfg_file = pathlib.Path(cfg)
        try:
            tree = ElementTree.parse(cfg_file)
            root = tree.getroot()

            package_data = {}
            for pkg in root.iter(section):
                for child in pkg:
                    package_data[child.tag] = [elem.text for elem in child.iter() if elem.text.strip()]

        except (FileNotFoundError, ElementTree.ParseError, AttributeError) as error:
            raise Ten8tException(f"XML config file {cfg} error: {error}") from error

        return package_data



"""Basic exception classes for ten8t."""


class Ten8tTypeError(TypeError):
    """Type errors associated with setting up test8t

    When bad types are sent to ten8t, this exception will be raised and should
    not be confused the TypeError that (should) indicate an unexpected lower
    level error.

    """


class Ten8tValueError(ValueError):
    """ Value Error associated with setting up test8t

    These exceptions will occur when setting up parameters on ten8t attributes and
    basic setup.  For example a negative weight.

    """


class Ten8tException(Exception):
    """Specialized exception for ten8t."""



"""
This module contains the `Ten8tPackage` class, which provides the functionality to load Ten8t
modules from a specified directory, manage these modules, and retrieve their properties.

The `Ten8tPackage` class provides the following main functionalities:

- Initialization allows specifying a directory, a naming glob pattern for the modules,
  a prefix for the functions, an option to autoload modules, a name for the package, and a
  dictionary of environment parameters.

- You can dynamically load modules from the directory matching the module_glob pattern.

- Access loaded modules directly through the `get` method and query the total
  number of modules via `module_count` property.

- Retrieve all the unique identifiers of rules existing within these modules through
  the `ruids` method.

The modules are added to the system path for easy import and use in other Python files.
"""
import pathlib
import sys

from .ten8t_exception import Ten8tException
from .ten8t_module import Ten8tModule
from .ten8t_result import Ten8tResult


class Ten8tPackage:
    """This class handles loading all modules from a folder that match a
       given file name pattern."""

    def __init__(
            self,
            folder="check",
            module_glob="check_*.py",
            function_prefix="check_",
            auto_load=True,
            name=None,
            env: dict | None = None,
    ):
        self.modules: list[Ten8tModule] = []
        self.folder: pathlib.Path = pathlib.Path(folder)
        self.module_glob: str = module_glob
        self.function_prefix: str = function_prefix
        self.env: dict = env or {}
        self.results: list[Ten8tResult] = []

        if not name:
            self.name = self.folder.name
        else:
            self.name = name

        self.folder = self.folder.resolve()

        # Add to module search path
        self._add_folder_to_sys_path(self.folder)

        self._verify_dir()

        if auto_load:
            self.load_modules()

    def _verify_dir(self):
        # Catastrophic, ten8t can't work with in check folder
        if not self.folder.exists():
            raise Ten8tException(
                f"The ten8t check folder '{self.folder}' does not exist."
            )

    @staticmethod
    def _add_folder_to_sys_path(folder):
        if not folder:
            return

        # Convert the folder path to an absolute path
        absolute_folder_path = pathlib.Path(folder).resolve()

        # Check if the absolute folder path is in sys.path
        if not any(
                pathlib.Path(path).resolve() == absolute_folder_path for path in sys.path
        ):
            # If it's not, add it to sys.path
            sys.path.insert(0, str(absolute_folder_path))

    @property
    def module_count(self) -> int:
        """ Count the number of modules in a package. """
        return 0 if not self.modules else len(self.modules)

    def load_modules(self, glob=None) -> list[Ten8tModule]:
        """ Find all the files that match the pattern and load the modules. """
        check_file_glob = glob or self.module_glob

        for file_path in sorted(self.folder.glob(check_file_glob)):
            module_name = f"{file_path.stem}"

            module = Ten8tModule(module_name, module_file=str(file_path), auto_load=True)
            self.modules.append(module)

        return self.modules

    def get(self, module_name) -> Ten8tModule | None:
        """Find a module given its name."""
        for module in self.modules:
            if module.module_name == module_name:
                return module
        return None

    def ruids(self):
        """get a list of all the RUIDS in a package"""
        ruids = []
        for module in self.modules:
            ruids.extend(module.ruids())
        return sorted(ruids)



"""
Attributes can be added to any Ten8t function using the @attributes decorator.

Attributes allow metadata to be added to rule functions to control how they are
run, filtered, and scored. In order to meet our minimalist sensibilities, we have
kept the number of attributes to a minimum and NONE are required in order to
minimize, nearly to zero the overhead of writing a rule.

This design philosophy matches a bit of the zen of python: "Simple is better than
complex." In order to write a simple test you are never required to add each and
every attribute to a rule. Defaults are provided for all attributes. You can go
along way never using an attribute...and once you learn them you will use them all
the time.
"""
import re

from .ten8t_exception import Ten8tException

DEFAULT_TAG = ""  # A string indicating the type of rule, used for grouping/filtering results
DEFAULT_LEVEL = 1  #
DEFAULT_PHASE = ""  # String indicating what phase of the dev process a rule is best suited for
DEFAULT_WEIGHT = 100  # The nominal weight for a rule should be a positive number
DEFAULT_SKIP = False  # Set to true to skip a rule
DEFAULT_TTL_MIN = 0  # Time to live for check functions.
DEFAULT_RUID = ""
DEFAULT_FINISH_ON_FAIL = False  # If a ten8t function yields fail result stop processing
DEFAULT_SKIP_ON_NONE = False
DEFAULT_FAIL_ON_NONE = False
DEFAULT_INDEX = 1  # All ten8t functions are given an index of 1 when created.
DEFAULT_THREAD_ID = "main_thread__"


def _parse_ttl_string(input_string: str) -> float:
    """
    Use regular expression to match a TTL string.  This pattern was a pain to figure out.  There
    are so many permutations that need to be handled that are subtle (like the order matters in the
    list).

    Args:
        input_string (str): The input string to parse.

    Returns:
        Tuple[Optional[float], Optional[str]]: A tuple containing the parsed floating-point number
        and optional units. Returns (None, None) if no match is found.
    """
    scale = {"seconds": 60,
             "second": 60,
             "sec": 60,
             "s": 60,
             "m": 1,
             "min": 1,
             "minute": 1,
             "minutes": 1,
             "h": 1 / 60.,
             "hr": 1 / 60.,
             "hrs": 1 / 60.,
             "hour": 1 / 60.}
    pattern = re.compile(
        r"([+-]?\d+\.\d*|\d*\.\d+|[-+]?\d+)\s*"
        r"(hour|hrs|hr|h|minutes|minute|min|m|seconds|second|sec|s)?"
    )
    matches = re.findall(pattern, input_string)
    if len(matches) == 1 and len(matches[0]) == 2:
        if matches[0][1] == '':
            unit = "m"
        else:
            unit = matches[0][1]
        number = float(matches[0][0]) / scale[unit]
        if number < 0.0:
            raise Ten8tException("TTL must be greater than or equal to 0.0")
        return number

    return 0.0


def attributes(
        *,
        tag=DEFAULT_TAG,
        phase=DEFAULT_PHASE,
        level=DEFAULT_LEVEL,
        weight=DEFAULT_WEIGHT,
        skip=DEFAULT_SKIP,
        ruid=DEFAULT_RUID,
        ttl_minutes=DEFAULT_TTL_MIN,
        finish_on_fail=DEFAULT_FINISH_ON_FAIL,  # Abort the whole run
        skip_on_none=DEFAULT_SKIP_ON_NONE,
        fail_on_none=DEFAULT_FAIL_ON_NONE,
        thread_id=DEFAULT_THREAD_ID,

):
    """
    Decorator to add attributes to a Ten8t function.

    Note the *, I always forget that this means that the function is kwarg only.
    """

    # throws exception on bad input
    ttl_minutes = _parse_ttl_string(str(ttl_minutes))

    if weight in [None, True, False] or weight <= 0:
        raise Ten8tException("Weight must be numeric and > than 0.0.  Nominal value is 100.0.")

    # Make sure these names don't have bad characters.  Very important for regular expressions
    disallowed = ' ,!@#$%^&:?*<>\\/(){}[]<>~`-+=\t\n\'"'
    for attr_name, attr in (('tag', tag), ('phase', phase), ('ruid', ruid)):
        bad_chars = [c for c in disallowed if c in attr]
        if bad_chars:
            raise Ten8tException(f"Invalid characters {bad_chars} found in {attr_name} ")

    def decorator(func):
        """Jam in all the attributes"""
        func.phase = phase
        func.tag = tag
        func.level = level
        func.weight = weight
        func.skip = skip
        func.ruid = ruid
        func.ttl_minutes = ttl_minutes
        func.finish_on_fail = finish_on_fail
        func.skip_on_none = skip_on_none
        func.fail_on_none = fail_on_none
        func.thread_id = thread_id
        return func

    return decorator


def get_attribute(func, attr, default_value=None):
    """
    Returns an attribute from a function.
    """
    defs = {
        "tag": DEFAULT_TAG,
        "phase": DEFAULT_PHASE,
        "level": DEFAULT_LEVEL,
        "weight": DEFAULT_WEIGHT,
        "skip": DEFAULT_SKIP,
        "ruid": DEFAULT_RUID,
        "ttl_minutes": DEFAULT_TTL_MIN,
        "finish_on_fail": DEFAULT_FINISH_ON_FAIL,
        "skip_on_none": DEFAULT_SKIP_ON_NONE,
        "fail_on_none": DEFAULT_FAIL_ON_NONE,
        "index": DEFAULT_INDEX,
        "thread_id": DEFAULT_THREAD_ID,
    }

    default = default_value or defs[attr]

    return getattr(func, attr, default)




""" Baseline rules for checking contents of Excel files. """
import openpyxl
import pandas as pd

from .ten8t_exception import Ten8tException
from .ten8t_format import BM
from .ten8t_result import TR
from .ten8t_util import str_to_bool

SHEET1 = "Sheet1"
AUTO = "auto"
DESC_DEFAULT = ""
START_ROW_DEFAULT = "1"
VAL_COL_DEFAULT = "B"
DESCRIPTION_COLUMN = "A"


def _column_to_number(column: str) -> int:
    number = 0
    for i, char in enumerate(reversed(column)):
        number += (ord(char.upper()) - 64) * (26 ** i)
    return number


def _get_sheet(wb, sheet_name=None):
    """Ensure a valid sheet is selected based on sheet_name parameter"""
    if sheet_name is None:
        return wb["Sheet1"]
    if sheet_name in wb.sheetnames:
        return wb[sheet_name]
    if len(wb.sheetnames) == 1:
        return wb[wb.sheetnames[0]]
    if "Sheet1" in wb.sheetnames:
        return wb["Sheet1"]

    raise Ten8tException('A sheet name was not specified and sheet1 could not be found.')


def _ensure_row_params(row_end, row_start: int):
    """Ensure the start and end rows parameters are consistent"""
    auto = False

    if row_end is None:
        return row_start, row_start, auto

    if isinstance(row_end, str):
        if row_end.isdigit() and int(row_end) < row_start:
            raise Ten8tException(
                f'Value for end row must be larger than start row {row_start=} {row_end=}')
        if row_end.lower() == AUTO:
            auto = True
            row_end = 1000
    try:
        row_end = int(row_end)
    except ValueError as vex:
        raise Ten8tException("row_end was not a valid integer value") from vex
    return row_start, row_end, auto


def rule_xlsx_a1_pass_fail(wb: openpyxl.workbook.Workbook,
                           sheet_name: str | None = None,
                           desc_col='A',
                           val_col='B',
                           row_start='1',
                           row_end=None):
    """ This is a very blunt instrument that pulls a true/false value out of
        a specific sheet/row/col of an Excel workbook.  It is very unforgiving
        to format changes in the work book

        Using start and end row numbers you can iterate over many items in the notebook.  If row
        end is set to 'auto' it will run until the first blank is detected in the value column.

        """
    sheet = _get_sheet(wb, sheet_name)

    # Handle Nones.  Presumably this should not be required
    row_start = row_start or '1'
    val_col = val_col or 'B'
    row_start, row_end, auto = _ensure_row_params(row_end, int(row_start))
    val_col = _column_to_number(val_col)

    if desc_col is not None:
        desc_col = _column_to_number(desc_col)

    for row in range(row_start, row_end + 1):

        value = sheet.cell(row=row, column=val_col).value
        if value is None and auto:
            break

        if value is None:
            raise Ten8tException(f'Expected boolean value in row {row}')

        if desc_col is not None:
            desc = sheet.cell(row=row, column=desc_col).value
        else:
            # It is possible not to have a description column
            desc = ""

        if str_to_bool(value):
            yield TR(status=True, msg=f"{BM.expected(desc)}-Passed")
        else:
            yield TR(status=False, msg=f"{BM.expected(desc)}-Failed")


def rule_xlsx_df_pass_fail(df: pd.DataFrame, desc_col: str, val_col: str, skip_on_none=False):
    """
    One could argue this is really a dataframe tool, but it is assumed that the end
    user will load an Excel spreadsheet into a dataframe and then process it.  This
    guy looks at two columns assuming the first row is the column header
    """

    for row in df.values:
        # Make dictionaries because they are easier to look at.
        row_dict = dict(zip(df.columns, row))

        if pd.isnull(row_dict[val_col]):
            if skip_on_none:
                yield TR(status=None, skipped=True,
                         msg=f"Null value detected in column={BM.expected(val_col)}")
            else:
                yield TR(status=False,
                         msg=f"Null value detected in column={BM.expected(val_col)}")
            continue

        if pd.isnull(row_dict[desc_col]):
            if skip_on_none:
                yield TR(status=None, skipped=True,
                         msg=f"Null description detected in column={BM.expected(desc_col)}")
            else:
                yield TR(status=False,
                         msg=f"Null description detected in column={BM.expected(desc_col)}")
            continue

        description = row_dict[desc_col]

        # Very lenient boolean values
        status = str_to_bool(row_dict[val_col])
        if status:
            yield TR(status=True, msg=f"{BM.code(description)}-Passed")
        else:
            yield TR(status=False, msg=f"{BM.code(description)}-Failed")



"""Basic scoring algorithms for Ten8t test results. """

import abc
from typing import Any

from .ten8t_exception import Ten8tException
from .ten8t_result import Ten8tResult


class ScoreStrategy(abc.ABC):
    """
    A strategy for scoring the results of a Ten8t run.
    It is assumed that many scoring strategies could be implemented, this provides a way
    for those strategies to be implemented in code (by providing a class) or from file
    by providing a name that matches the class strategy name attribute.
    """

    strategy_name: str | None = None

    @abc.abstractmethod
    def score(self, results: list[Ten8tResult]) -> float:  # pragma: no cover
        """Abstract score method"""

    def __call__(self, results: list[Ten8tResult]):
        return self.score(results)

    @classmethod
    def strategy_factory(cls, strategy_name_or_class) -> "ScoreStrategy":
        """Make a strategy object from a name or class.
        This will be read from files or code, so they support both"""
        if isinstance(strategy_name_or_class, str):
            # Note this only goes one level deep in subclasses, which for now is good enough.
            for subclass in cls.__subclasses__():
                if strategy_name_or_class in (subclass.strategy_name, subclass.__name__):
                    return subclass()
            raise Ten8tException(
                f"No scoring strategy with name '{strategy_name_or_class}' found."
            )
        if issubclass(strategy_name_or_class, cls):
            return strategy_name_or_class()
        raise Ten8tException(
            "Argument must be a strategy name or a ScoreStrategy subclass.")


class ScoreByResult(ScoreStrategy):
    """Calculate the score by individually weighting each result"""

    strategy_name = "by_result"

    def score(self, results: list[Ten8tResult] | None = None) -> float:
        """Weighted result of all results."""

        if results is None:
            return 0.0

        weight_sum = 0.0
        passed_sum = 0.0
        # Remove any skipped results
        results = [result for result in results if not result.skipped]
        if not results:
            return 0.0
            # raise Ten8tException("No results to score.")

        for result in results:
            passed_sum += result.weight if result.status else 0.0
            weight_sum += result.weight

        return (100.0 * passed_sum) / (weight_sum * 1.0)


class ScoreByFunctionBinary(ScoreStrategy):
    """Calculate the score by requiring ALL results from a function
       to have status=True to consider the function passed."""

    strategy_name = "by_function_binary"

    def score(self, results: list[Ten8tResult] | None = None) -> float:
        """If any result on a function fails then the function fails."""
        if results is None:
            return 0.0

        score_functions: dict[str, Any] = {}

        for result in results:
            key = f"{result.pkg_name}.{result.module_name}.{result.func_name}".lstrip(
                "."
            )
            score_functions.setdefault(key, []).append(result)

        # Remove any skipped results
        results = [result for result in results if not result.skipped]
        if not results:
            return 0.0

        for key, results_ in score_functions.items():
            if not results_:
                score_functions[key] = 0.0
            else:
                score_functions[key] = 100.0 if all(r.status for r in results_) else 0.0

        # The score should be the average of the scores for each function
        return sum(score_functions.values()) / (len(score_functions) * 1.0)


class ScoreByFunctionMean(ScoreStrategy):
    """Calculate score by averaging the results from a function.
       This means that a function could 50% pass"""

    strategy_name = "by_function_mean"

    def score(self, results: list[Ten8tResult] | None = None) -> float:
        """Find the average of the results from each function."""
        if results is None:
            return 0.0

        function_results: dict[str, Any] = {}

        # Remove any skipped results
        results = [result for result in results if not result.skipped]
        if not results:
            return 0.0

        for result in results:
            key = f"{result.pkg_name}.{result.module_name}.{result.func_name}".lstrip(
                "."
            )
            function_results.setdefault(key, []).append(result)

        sum_weights = 0.0
        sum_passed = 0.0

        # Now we have a dictionary of results for each function.  We can now score each function
        for key, results_ in function_results.items():
            for result in results_:
                sum_weights += result.weight
                sum_passed += result.weight if result.status else 0.0

        # This does not appear to be possible.  The empty list is protected against
        # and each of the summed weights must be > 0.  This could be removed?
        if sum_weights == 0.0:
            raise Ten8tException("The sum of weights is 0.  This is not allowed.")

        # The score should be the average of the scores for each function
        return (100.0 * sum_passed) / (sum_weights * 1.0)


class ScoreBinaryFail(ScoreStrategy):
    """Anything fails then the test is a fail.  Empty results fail."""

    strategy_name = "by_binary_fail"

    def score(self, results: list[Ten8tResult] | None) -> float:
        if results is None:
            return 0.0

        if not results:
            return 0.0

        if any(not result.status for result in results if not result.skipped):
            return 0.0
        return 100.0


class ScoreBinaryPass(ScoreStrategy):
    """Anything passes then the test is a pass. Empty results fail. """
    strategy_name = "by_binary_pass"

    def score(self, results: list[Ten8tResult] | None) -> float:
        if results is None:
            return 0.0

        if any(result.status for result in results if not result.skipped):
            return 100.0
        return 0.0



"""
This module contains the rules for interacting with dataframes.  Instead of using a single type of
dataframe like pandas, polars or versions of these data frames, I'm using narwhals which is
a compatibility layer that allows ANY type of dataframe to be handled.
"""
from typing import Generator

import narwhals as nw
from narwhals.typing import FrameT

from .ten8t_exception import Ten8tException
from .ten8t_format import BM
from .ten8t_result import TR, Ten8tYield
from .ten8t_util import any_to_str_list


@nw.narwhalify()
def rule_validate_ndf_schema(df: FrameT,
                             int_cols: str | list[str] = None,
                             float_cols: str | list[str] = None,
                             str_cols: str | list[str] = None,
                             number_cols: str | list[str] = None,
                             no_null_cols: str | list[str] = None,
                             summary_name: str | None = '',
                             summary_only: bool = False,
                             name=None) -> Generator[TR, None, None]:
    """
        Validates the schema of a DataFrame according to the specified rules.

        Parameters:
        df: The DataFrame to validate
        int_cols: A list or a string representing column(s) that should contain integers
        float_cols: A list or a string representing column(s) that should contain floats
        str_cols: A list or a string representing column(s) that should contain strings
        number_cols: A list or a string representing column(s) that should contain numeric values (either int or float)
        no_null_cols: A list or a string representing column(s) that should not contain null values
        summary_name: An optional name for the summary of the validation result
        summary_only: A boolean flag indicating whether to yield only the summary of the validation result
        name: An optional name for this rule

        Returns:
        A generator of TR (Test Result) objects, each representing the validation result of a particular rule on a column

        Raises:
        Ten8tException: If no schema checks (i.e., no columns for each of the column types) were specified
    """
    int_cols = any_to_str_list(int_cols)
    float_cols = any_to_str_list(float_cols)
    number_cols = any_to_str_list(number_cols)
    str_cols = any_to_str_list(str_cols)
    no_null_cols = any_to_str_list(no_null_cols)
    name = name or "rule_validate_ndf_schema"
    summary_name = summary_name or "data frame."

    if not any((int_cols, float_cols, no_null_cols, str_cols, number_cols)):
        raise Ten8tException(f"No schema checks were specified for {name}")

    schema = df.dtypes

    y = Ten8tYield(summary_name=summary_name, summary_only=summary_only)

    int_types = "int8 int16 int32 int64 uint8 uint16 uint32 uint64".split()
    float_types = "float32 float64".split()
    num_types = int_types + float_types
    str_types = "object".split()

    for int_col in int_cols:
        if schema[int_col].name in int_types:
            yield from y(TR(True, f"Column {int_col} is an integer type"))
        else:
            yield from y(TR(False, f"Column {BM.code(int_col)} is {BM.bold('NOT')} an integer type"))

    for float_col in float_cols:
        if schema[float_col].name in float_types:
            yield from y(TR(True, f"Column {BM.code(float_col)} is a float type"))
        else:
            yield from y(TR(False, f"Column {BM.code(float_col)} is {BM.bold('NOT')} an float type"))

    for number_col in number_cols:
        if schema[number_col].name in num_types:
            yield from y(TR(True, f"Column {BM.code(number_col)} is a number type"))
        else:
            yield from y(TR(False, f"Column {BM.code(number_col)} is {BM.bold('NOT')} an number type"))

    for str_col in str_cols:
        if schema[str_col].name in str_types:
            yield from y(TR(True, f"Column {BM.code(str_col)} is an string type"))
        else:
            yield from y(TR(False, f"Column {BM.code(str_col)} is {BM.bold('NOT')} an string type"))

    for no_null_col in no_null_cols:
        null_count = len(v for v in df[no_null_col] if v is None or v == '')
        if null_count == 0:
            yield from y(TR(True, f"Column {BM.code(no_null_col)} has {null_count} empty/null values."))
        else:
            yield from y(TR(False, f"Column {BM.code(no_null_col)} has {null_count} empty/null values"))


@nw.narwhalify()
def rule_ndf_columns_check(name: str, df: FrameT, expected_cols_: str | list[str], exact=False):
    """
    A FrameT is a narwhals dataframe.  This supports ANY version of pandas, polars etc. based on what you have
    pip installed into your project.  This allows us to not worry about which version you have installed.

    NOTE: narwhalify takes the dataframe that you pass in whatever type it happens to be and converts
          it to a narwhals dataframe.  This conversion is cheap in that it is just a function wrapper.

    Args:
        name: The name of the dataframe, used for reporting
        df: The dataframe to check
        expected_cols: A list or space separated string of column names that are expected in the dataframe
        exact: If true, the dataframe must have exactly the expected columns. 
               If false, the dataframe can have additional columns.

    Returns:
        A generator that yields Ten8tResult object(s) based on the column checks.
    """

    if df.is_empty() or not df.columns:
        if not expected_cols_:
            yield TR(status=True, msg=f"The {BM.code(name)} data frame is empty and there are no expected columns.")
        else:
            raise ValueError(f"There are no columns in {BM.code(name)}.")

    df_col_names = set(df.columns)
    expected_cols = set(any_to_str_list(expected_cols_))

    missing_columns = expected_cols - df_col_names
    extra_columns = df_col_names - expected_cols

    if exact:
        if not missing_columns and not extra_columns:
            yield TR(status=True,
                     msg=f"All columns were found in the {BM.code(name)} dataframe: {BM.expected(expected_cols)}")
        else:
            msg = f"Data frame {BM.code(name)} is missing {BM.expected(missing_columns)} columns and has extra {BM.expected(extra_columns)} columns"
            yield TR(status=False, msg=msg)

    else:
        # The not exact case means there are no missing but may be extra.
        if not missing_columns:
            yield TR(status=True,
                     msg=f"Data frame {BM.code(name)} has NO missing columns")
        else:
            yield TR(status=False,
                     msg=f"Data frame {BM.code(name)} has missing columns {BM.expected(missing_columns)}")


def convert_to_tuple(input_val: tuple[float, list[str] | str] | None) -> tuple[float, list[str]] | None:
    """
    Convert data inf the form (1.23,[1,2]) and (1.23,"1 2") into a tuple
    with the values (1.23,[1,2]) while keeping mypy happy.
    """
    if input_val is None:
        return None
    value, arr = input_val
    if isinstance(arr, str):
        arr = tuple(float(x) for x in arr.replace(',', ' ').split())
    return value, arr


@nw.narwhalify()
def rule_validate_ndf_values_by_col(df: FrameT,
                                    positive: list[str] | str | None = None,
                                    non_negative: list[str] | str | None = None,
                                    percent: list[str] | str | None = None,
                                    min_: tuple[float, list[str]] | None = None,
                                    max_: tuple[float, list[str]] | None = None,
                                    negative: list[str] | str | None = None,
                                    non_positive: list[str] | str | None = None,
                                    correlation: list[str] | str | None = None,
                                    probability: list[str] | str | None = None,
                                    name: str = '') -> Generator[TR, None, None]:
    """
        Validate DataFrame schema based on given conditions. Parameters are as follows:

        Parameters:
        - df (pd.DataFrame): DataFrame to validate.
        - columns (Sequence[str], optional): Col names to validate. Defaults to None.
        - no_null_columns (Sequence[str], optional): Cols shouldn't have null values. Defaults to None.
        - int_columns (Sequence[str], optional): Columns of integer type. Defaults to None.
        - float_columns (Sequence[str], optional): Columns of float type. Defaults to None.
        - str_columns (Sequence[str], optional): Columns of string type. Defaults to None.
        - row_min_max (Tuple[int,int], optional): Min/max # of rows in DataFrame. Defaults to None.
        - allowed_values (Sequence, optional): Allowed values in columns. Defaults to None.
        - empty_ok (bool, optional): If True, empty DataFrame is valid. Defaults to False.

        Raises:
        - Ten8tException: If df is None, min_rows/max_rows aren't positive integers, or
                           min_rows is > max_rows.

        Yields:
        - TR: An obj with the status of the validation (True if condition is met, False otherwise)
              and a message describing the result.
    """

    if df is None or df.is_empty():
        raise Ten8tException("Dataframe is empty or None")

    # THis is probably the most readable way to do this...but using vars['variable'] is tempting
    positive = any_to_str_list(positive, sep=',')
    non_negative = any_to_str_list(non_negative, sep=',')
    percent = any_to_str_list(percent, sep=',')
    negative = any_to_str_list(negative, sep=',')
    non_positive = any_to_str_list(non_positive, sep=',')
    correlation = any_to_str_list(correlation, sep=',')
    probability = any_to_str_list(probability, sep=',')

    if min_:
        min_ = convert_to_tuple(min_)
    if max_:
        max_ = convert_to_tuple(max_)

    conditions = [positive, non_negative, negative, non_positive,
                  percent, min_, max_, probability, correlation]

    if not any(conditions):
        raise Ten8tException("No data frame column value rules specified.")

    if positive:
        for col in positive:
            if df[col].min() > 0.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are positive.")

            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')}  positive.")

    if non_negative:
        for col in non_negative:
            if df[col].min() >= 0.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are non-negative.")
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} non-negative.")

    if percent:
        # Just check 0-100
        for col in percent:
            if 0.0 <= df[col].min() and df[col].max() <= 100.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are  a percent.")
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} a percent.")

    if probability:
        for col in probability:
            # Just check 0-100
            if 0.0 <= df[col].min() and df[col].max() <= 1.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are probabilities.", )
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} probabilities.")

    if correlation:
        for col in correlation:
            # Just check 0-100
            if -1.0 <= df[col].min() and df[col].max() <= 1.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are correlations.", )
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} correlations.")

    if min_:
        val, cols = min_
        for col in cols:
            if df[col].min() >= val:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are > {val}")
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} > {val}")

    if max_:
        val, cols = max_
        for col in cols:
            if df[col].max() <= val:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are < {val}")
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} < {val}")

    if negative:
        for col in negative:
            if df[col].max() < 0.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are negative.")
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} negative.")

    if non_positive:
        for col in non_positive:
            if df[col].max() <= 0.0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are non-positive.")
            else:
                yield TR(status=False, msg=f"All values in {BM.code(col)} are {BM.bold('NOT')} non-positive.")

    if non_positive:
        for col in non_positive:
            if df[col].max() < 0:
                yield TR(status=True, msg=f"All values in {BM.code(col)} are non-positive.")
            else:
                yield TR(status=False, msg=f"All values in {col} are {BM.bold('NOT')} non-positive.")


def extended_bool(value) -> bool:
    """
    Spreadsheet friendly boolean values
    Args:
        value: string

    Returns:

    """
    truthy_values = [True, 1, '1', 'true', 't', 'pass', 'p', 'yes', 'y']
    untruthy_values = [None, False, 0, '0', 'false', 't' 'fail', 'f', 'no', 'n']

    # If value is boolean True, return True
    if isinstance(value, bool) and value is True:
        return True

    # For non-boolean values, convert to string and do a case-insensitive comparison
    str_value = str(value).lower()
    if str_value in truthy_values:
        return True
    elif str_value in untruthy_values:
        return False

    # This could be an exception
    return False


@nw.narwhalify()
def rule_ndf_pf_columns(df: FrameT,
                        pf_col: str = "Status",
                        desc_col: str = "Description",
                        enabled_col: str = None,
                        name: str = '',
                        summary_only: bool = False) -> Generator[TR, None, None]:
    """
    Given a dataframe of the form:
    Description  Status  Enabled
    Test1        1       1
    Test2        0       1
    Test3        0       0
    
    Check the status honoring the enabled column.  The enable column is optional, but it is assumed
    that this could come from a 'live' spreadsheet, so we need to be flexible on what "pass" means.
    
    In this case it can be True, 1, "Pass"
    
    Args:
        df: Data Frame
        pf_col: Pass Fail Column
        desc_col:  Description Column
        enabled_col: Enabled Column (when not included assumes all are enabled)
        name: Name used for summary
        summary_only: Only display summary?

    Returns:

    """
    if not desc_col:
        raise AttributeError("Invalid description column")
    if not pf_col:
        raise AttributeError("Invalid pass/fail column")
    if isinstance(enabled_col, str) and enabled_col.strip() == '':
        raise AttributeError("Invalid enable column name.")

    summary_yield = Ten8tYield(summary_only=summary_only)

    for row in df.iter_rows(named=True):
        if enabled_col and not extended_bool(row[enabled_col]):
            continue

        pass_status = extended_bool(row[pf_col])
        row_description = BM.code(row[desc_col])
        ten8t_result = TR(pass_status,
                          f"{row_description} {BM.pass_('passed.') if pass_status else BM.fail('failed.')}")

        yield from summary_yield(ten8t_result)

    if summary_only:
        yield from summary_yield.yield_summary(name=name)



"""
Set of baseline rules that uses the pyfilesystem module to OS-agnostic checks on things
about the file system, file existing, age etc.
"""
import datetime as dt
import fnmatch
from typing import Generator, Sequence

import humanize
from fs.base import FS
from fs.errors import FSError
from fs.osfs import OSFS

from .ten8t_format import BM
from .ten8t_result import TR


def rule_fs_paths_exist(fs_obj: OSFS, paths: Sequence[str]) -> Generator[TR, None, None]:
    """ Check a bunch of paths."""
    for path in paths:
        yield from rule_fs_path_exists(fs_obj, path)


def rule_fs_path_exists(fs_obj: OSFS, path_: str) -> Generator[TR, None, None]:
    """Simple rule to check for a file path."""
    yield TR(status=fs_obj.exists(path_), msg=f"The path {path_} on {fs_obj.root_path} exists.")


def human_readable_size(size_in_bytes: int, binary=True) -> str:
    """ Dump bytes in human-readable form"""
    return humanize.naturalsize(float(size_in_bytes), binary=binary)


def rule_fs_file_within_max_size(filesys: OSFS,
                                 path: str,
                                 max_file_size: int,
                                 binary: bool = False,
                                 skip_if_missing=False):
    """Check if a file exists and its size is within the given max_file_size limit"""
    if not filesys.isfile(path):
        yield TR(status=False, msg=f'File "{path}" does {BM.bold("NOT")} exist in {filesys.root_path}',
                 skipped=skip_if_missing)
    else:
        file_size = filesys.getsize(path)
        file_size_str = human_readable_size(file_size, binary=binary)
        delta = file_size - max_file_size
        # delta_str = human_readable_size(abs(delta))
        if delta < 0:
            yield TR(status=False, msg=f'File "{path}" size={file_size_str} exceeds size limit by -{delta} bytes.')
        else:
            yield TR(status=True, msg=f'File "{path}" size={file_size_str} within size limit by {delta} bytes.')


def sec_format(seconds):
    """
    Convert a given time in seconds to a string expressing the time in a more
    human-readable format.  Of note when calculating months it uses 30-day months which is fine
    for a few months but at 23 months it has 10 days of error.

    Time is rounded down in the largest time units possible (days, hours, minutes, then seconds)
    For time less than 2 days, it's represented in hours. For less than 2 hours, it's in minutes.
    For less than 2 minutes, it's in seconds. The seconds are displayed with up to three
    digits of precision.

    NOTE: Months are problematic since they aren't constantly sized (same with years), I'm banking
          on this being human-readable and that very slight errors  in the last decimal point
          are a problem.

    Parameters:
    seconds (float): The time duration in seconds.

    Returns:
    string: A string representation of the provided time duration making it easier to read.

    """
    seconds_per_minute = 60
    seconds_per_hour = 60 * seconds_per_minute
    seconds_per_day = 24 * seconds_per_hour
    seconds_per_month = 30 * seconds_per_day  # Human-readable, don't whine about this
    seconds_per_year = 365 * seconds_per_day

    # Check if seconds is very small negative, if yes then round it to 0
    if -0.001 < seconds < 0:
        seconds = 0

    # Store the original sign and use absolute value for calculations.
    sign = '-' if seconds < 0 else ''
    seconds = round(abs(seconds), 3)

    if seconds == 0:
        sign = ''

    # Order of items matters, biggest goes first
    time_units = [
        ("years", seconds_per_year),
        ("months", seconds_per_month),
        ("days", seconds_per_day),
        ("hours", seconds_per_hour),
        ("minutes", seconds_per_minute)
    ]

    for unit, sec_in_unit in time_units:
        if seconds >= 2 * sec_in_unit:
            return f"{sign}{seconds / sec_in_unit:.1f} {unit}"

    # Seconds are special case. For small seconds we report to
    # extra precision since we assume that we are dealing with
    # times on a human scale rather than on micro scales.
    if seconds >= 2:
        return f"{sign}{seconds:.1f} seconds"

    if seconds == 0:
        return "0.000 seconds"  # no sign

    return f'{sign}{seconds:.3f} seconds'


def rule_fs_oldest_file_age(filesys: FS, max_age_minutes: float = 0,
                            max_age_hours: float = 0,
                            max_age_days: float = 0,
                            max_age_seconds: float = 0,
                            patterns=None,
                            no_files_stat=True,
                            no_files_skip=True,
                            now__: dt.datetime | None = None):
    """
    This rule is useful for ensuring that files are being removed in
    a timely manner as in the case where a folder is used to queue up
    data files that are processed.  Old files indicates an issue.
    """
    patterns = patterns or ['*']

    if isinstance(patterns, str):
        patterns = patterns.split(',')

    now_: dt.datetime = (now__ or dt.datetime.now(dt.timezone.utc)).replace(tzinfo=dt.timezone.utc)
    max_file_age_seconds = dt.timedelta(days=max_age_days,
                                        hours=max_age_hours,
                                        minutes=max_age_minutes,
                                        seconds=max_age_seconds).total_seconds()

    try:
        files = filesys.listdir('/')
        files = [f for f in files if
                 filesys.isfile(f) and any(
                     fnmatch.fnmatch(f, pattern) for pattern in patterns)]
    except FSError as e:
        yield TR(status=False, msg=f"Error during listing files: {str(e)}", except_=e)
        return

    if not files:
        yield TR(status=no_files_stat,
                 msg=f"No files found in the directory: {filesys.getsyspath('/')}",
                 skipped=no_files_skip)
        return

    try:
        oldest_file = min(files, key=lambda f: filesys.getinfo(f, namespaces=['details']).modified)
        oldest_file_modified: dt.datetime = filesys.getinfo(oldest_file, namespaces=['details']).modified
        oldest_file_age_seconds = (now_ - oldest_file_modified).total_seconds()
    except FSError as e:
        yield TR(status=False, msg=f"Error during checking file's age: {str(e)}", except_=e)
        return

    time_str = sec_format(max_file_age_seconds)
    old_str = sec_format(oldest_file_age_seconds)

    if oldest_file_age_seconds <= max_file_age_seconds:
        yield TR(status=True,
                 msg=f'Oldest file "{oldest_file}" is within age limit of {time_str}. File age= {old_str}')
    else:
        yield TR(status=False,
                 msg=f'Oldest file "{oldest_file}" is more than {time_str}. File age= {old_str}')



""" This module contains the Ten8tResult class and some common result transformers. """

import itertools
import traceback
from collections import Counter
from dataclasses import asdict, dataclass, field
from functools import wraps
from operator import attrgetter
from typing import Any, Generator, Sequence

from .ten8t_exception import Ten8tException
from .ten8t_format import Ten8tMarkup


@dataclass
class Ten8tResult:
    """
    Return value of a Ten8tFunction.

    This dataclass tracks the status of a Ten8tFunction. It includes data relating to the function
    call, such as the status, module name, function name, message, additional info, warning message,
    docstring, runtime, exceptions, traceback, skip flag, tag, level, and count.

    This data can be used for reporting purposes.

    Attributes:
        status (bool): Check status. Default is False.
        module_name (str): Module name. Default is "".
        func_name (str): Function name. Default is "".
        msg (str): Message to the user. Default is "".
        info_msg (str): Additional function call info. Default is "".
        warn_msg (str): Warning message. Default is "".
        doc (str): Function docstring. Default is "".
        runtime_sec (float): Function runtime in seconds. Default is 0.0.
        except_ (Exception): Raised exception, if any. Default is None.
        traceback (str): Exception traceback, if any. Default is "".
        skipped (bool): Function skip flag. Default is False.
        tag (str): Function tag. Default is "".
        level (int): Function level. Default is 1.
        thread_id(str):Thread the function ran on. Default is "".
        count (int): Return value count from a Ten8tFunction.
        """

    status: bool | None = False

    # Name hierarchy
    func_name: str = ""
    pkg_name: str = ""
    module_name: str = ""

    # Msg Hierarchy
    msg: str = ""
    info_msg: str = ""
    warn_msg: str = ""

    msg_rendered = ""

    # Function Info
    doc: str = ""

    # Timing Info
    runtime_sec: float = 0.0

    # Error Info
    except_: Exception | None = None
    traceback: str = ""
    skipped: bool = False

    weight: float = 100.0

    # Attribute Info - This needs to be factored out?
    tag: str = ""
    level: int = 1
    phase: str = ""
    count: int = 0
    ruid: str = ""
    ttl_minutes: float = 0.0

    # Mitigations
    mit_msg: str = ""
    owner_list: list[str] = field(default_factory=list)

    # Bad parameters
    skip_on_none: bool = False
    fail_on_none: bool = False

    # Indicate summary results, so they can be filtered
    summary_result: bool = False

    # Thread id where function ran
    thread_id: str = ""

    mu = Ten8tMarkup()

    def __post_init__(self):
        # Automatically grab the traceback for better debugging.
        if self.except_ is not None and not self.traceback:
            self.traceback = traceback.format_exc()

    def as_dict(self):
        """Convert the Ten8tResult instance to a dictionary."""
        d = asdict(self)
        d['except_'] = str(d['except_'])
        return d


# Shorthand
TR = Ten8tResult


class Ten8tYield:
    """
    This allows syntactic sugar to know how many times a generator
    has been fired and how many passes and fails have occurred.

    These internal counts allow top level code to NOT manage that
    state at the rule level.  Instead, you just report your passes

    and fails and ask at the end how it played out.

    gen = BrickYield()

    if cond:
        yield from gen(BR(True,"Info...")
    if not gen.yielded:
        yield from gen(BR(False,"Nothing to do"))
    if show_summary:
        yield BR(status=self.fail_count==0,msg=f"{self.pass_count} passes and {self.fail_count} fails")

    """

    def __init__(self, summary_only=False, summary_name=""):
        """
        The ten8t yield class allows you to use the yield mechanism while also tracking
        pass fail status of the generator.  Using this class allows for a separation of
        concerns so your top level code doesn't end up counting passes and fails.
        
        When your test is complete you can query the yield object and report that
        statistics without a bunch of overhead.
        
        If you set summary_only to true, no messages will be yielded, but you 
        can yield the summary message manually when you are done with the test.
        
        If you provide a name to this init then a generic summary message can be
        generated like this:
        
        y = Ten8tYield("Generic Test")
        y(status=True,msg="Test1")
        y(status=True,msg="Test2")
        y(status=False,msg="Test3")
        y.yield_summary()
        
        BR(status=False,msg="Generic Test had 2 pass and 1 fail results for 66.7%.")
        
        Args:
            summary_only(bool): Defaults to False
            summary_name: Defaults to ""
        """
        self._count = 0
        self._fail_count = 0
        self.summary_only = summary_only
        self.summary_name = summary_name

    @property
    def yielded(self):
        """ Have we yielded once?"""
        return self._count > 0

    @property
    def count(self):
        """How many times have we yielded?"""
        return self._count

    @property
    def fail_count(self):
        """How many fails have there been"""
        return self._fail_count

    @property
    def pass_count(self):
        """How many passes have there been"""
        return self.count - self._fail_count

    @property
    def counts(self):
        """Return pass/fail/total yield counts"""
        return self.pass_count, self.fail_count, self.count

    def increment_counter(self, result: Ten8tResult) -> None:
        self._count += 1
        if not result.status:
            self._fail_count += 1

    def results(self, results: Ten8tResult | list[Ten8tResult]) -> Generator[Ten8tResult, None, None]:
        """
        This lets you pass a result or results to be yielded and mimics the way ten8t results
        work in other places where traditional result collection is used, for example code
        that returns a list of Ten8tResults
        Args:
            results: one or list of ten8t results
        Returns:

        """
        if isinstance(results, Ten8tResult):
            results = [results]

        if isinstance(results, Generator) or (isinstance(results, list) and isinstance(results[0], Ten8tResult)):
            # At this point we are iterating over a list or a generator.
            for result in results:
                if isinstance(result, Ten8tResult):
                    self.increment_counter(result)
                    if not self.summary_only:
                        yield result
                else:
                    raise Ten8tException(f"Unknown result type {type(results)}")
        else:
            raise Ten8tException(f"Unknown result type {type(results)}")

    def __call__(self, *args_, **kwargs_) -> Generator[Ten8tResult, None, None]:
        """
        Syntactic sugar to make yielding look just like creating the BR object at each
        invocation of yield.  The code mimics creating a Ten8tResult manually
        since the *args/**kwargs are passed through via a functools.wrapper. 
        
        y.results(BR(status=True,msg="Did it work?"))
        
        The __call_ override allows the following code to work correctly without having to manually
        instantiate a Ten8tResult. 
        
        y(status=True,msg="Did it work?")
        
        Under the covers all the parameters to this function are forward to the creation of
        the underlying Ten8tResult inside the wrapper.
        
                
        Args:
            *args_: For Ten8tResult 
            **kwargs_: For Ten8tResult
        """

        @wraps(Ten8tResult.__init__)
        def sr_wrapper(*args, **kwargs):
            """
            Make the __call__ method have the same parameter list as the Ten8tResult.__init__
            method.
            
            You can say:
            y(status=True,msg="Did it work?")
            
            or you can do
            
            y(BR(status=True,msg="Did it work?")
            
            Args:
                *args:   Handle any function args
                **kwargs: Handle any function kwargs

            Returns:

            """
            return Ten8tResult(*args, **kwargs)

        # If they just hand you a result then just pass it on
        if len(args_) == 1 and len(kwargs_) == 0 and isinstance(args_[0], Ten8tResult):
            results = [args_[0]]

        # This is when we get a generator
        elif len(args_) == 1 and len(kwargs_) == 0 and isinstance(args_[0], Generator):
            results = (x for x in args_[0])
        else:
            # package up the result information
            results = [sr_wrapper(*args_, **kwargs_)]
        for result in results:
            self.increment_counter(result)
            if not self.summary_only:
                yield result

    def yield_summary(self, name="", msg="") -> Generator[Ten8tResult, None, None]:
        """
        The yield summary should be the name of the summary followed information message
        about the summary.  The message should give a pass and fail count.  If no name
        or message is provided the function name is used and a generic message is
        created. Generally the name should be provided since the function name is only
        good enough for very simple cases.  In general the message is good enough since
        it is nice to have all summaries look the same with the pass and fail count.
        
        Since this is yielding a summary the summary_result flag is set to enable filtering.
        Args:
            name: 
            msg: 

        Returns:

        """
        name = name or self.summary_name or self.__call__.__name__
        msg = msg or f"{name} had {self.pass_count} pass and {self.fail_count} fail."

        yield Ten8tResult(status=self.fail_count == 0, msg=msg, summary_result=True)


# Result transformers do one of three things, nothing and pass the result on, modify the result
# or return None to indicate that the result should be dropped.  What follows are some
# common result transformers.

def passes_only(sr: Ten8tResult):
    """ Return only results that have pass status"""
    return sr if sr.status else None


def fails_only(sr: Ten8tResult):
    """Filters out successful results.

    Args:
        sr (Ten8tResult): The result to check.

    Returns:
        Ten8tResult: The result if it has failed, otherwise None.
    """
    return None if sr.status else sr


def remove_info(sr: Ten8tResult):
    """Filter out messages tagged as informational

    Args:
        sr (Ten8tResult): The result to check.

    Returns:
        Ten8tResult: The result if it has failed, otherwise None.
    """
    return None if sr.info_msg else sr


def warn_as_fail(sr: Ten8tResult):
    """Treats results with a warning message as failures.

    Args:
        sr (Ten8tResult): The result to check.

    Returns:
        Ten8tResult: The result with its status set to False if there's a warning message.
    """
    if sr.warn_msg:
        sr.status = False
    return sr


def results_as_dict(results: list[Ten8tResult]):
    """Converts a list of Ten8tResult to a list of dictionaries.

    Args:
        results (list[Ten8tResult]): The list of results to convert.

    Returns:
        list[Dict]: The list of dictionaries.
    """
    return [result.as_dict() for result in results]


def group_by(results: Sequence[Ten8tResult], keys: Sequence[str]) -> dict[str, Any]:
    """
    Groups a list of Ten8tResult by a list of keys.

    This function allows for arbitrary grouping of Ten8tResult using the keys of the
    Ten8tResult as the grouping criteria.  You can group in any order or depth with
    any number of keys.

    Args:
        results (Sequence[Ten8tResult]): The list of results to group.
        keys (Sequence[str]): The list of keys to group by.S

    """

    if not keys:
        raise Ten8tException("Empty key list for grouping results.")

    key = keys[0]
    key_func = attrgetter(key)

    # I do not believe this is an actual test case as it would require a bug in
    # the code.  I'm leaving it here for now.
    # if not all(hasattr(x, key) for x in results):
    #    raise ten8t.Ten8tValueError(f"All objects must have an attribute '{key}'")

    # Sort and group by the first key
    results = sorted(results, key=key_func)
    group_results: list[tuple[str, Any]] = [(k, list(g)) for k, g in itertools.groupby(results, key=key_func)]

    # Recursively group by the remaining keys
    if len(keys) > 1:
        for i, (k, group) in enumerate(group_results):
            group_results[i] = (k, group_by(group, keys[1:]))

    return dict(group_results)


def overview(results: list[Ten8tResult]) -> str:
    """
    Returns an overview of the results.

    Args:
        results (list[Ten8tResult]): The list of results to summarize.

    Returns:
        str: A summary of the results.
    """

    result_counter = Counter(
        'skip' if result.skipped else
        'error' if result.except_ else
        'fail' if not result.status else
        'warn' if result.warn_msg else
        'pass'
        for result in results
    )

    total = len(results)
    passed = result_counter['pass']
    failed = result_counter['fail']
    errors = result_counter['error']
    skipped = result_counter['skip']
    warned = result_counter['warn']

    return f"Total: {total}, Passed: {passed}, Failed: {failed}, " \
           f"Errors: {errors}, Skipped: {skipped}, Warned: {warned}"



"""
Handles configuration abstraction for ten8t, includes classes to parse TOML and JSON.
"""
import re
from typing import Sequence

from .ten8t_exception import Ten8tException


class Ten8tRC:
    """
    Loads configurations from a dictionary.
    """

    def __init__(self, *, rc_d: dict | None = None):

        # Any falsy value ends up being valid...which means that setting
        # rcd=0 will work.
        rc_d = rc_d or {}

        # This gives a nicer error than what might happen when you pass randomness
        # to the next function
        if not isinstance(rc_d, dict):
            raise Ten8tException(f"Ten8tRC expects a dictionary but got '{type(rc_d)}'")

        # This is being quite paranoid, but allows for the user to only specify what is
        # needed rather that having data structures filled with []
        rc_data = {
            'display_name': rc_d.get('display_name', 'NoName'),
            'ruids': rc_d.get('ruids', []),
            'tags': rc_d.get('tags', []),
            'phases': rc_d.get('phases', []),
            'levels': rc_d.get('levels', [])
        }

        # These will get overwritten
        self.ruids: Sequence[str] = []
        self.ex_ruids: Sequence[str] = []
        self.phases: Sequence[str] = []
        self.ex_phases: Sequence[str] = []
        self.tags: Sequence[str] = []
        self.ex_tags: Sequence[str] = []
        self.levels: Sequence[int] = []
        self.ex_levels: Sequence[int] = []

        self.expand_attributes(rc_data)
        self.name = rc_data['display_name']

        # This is a special case
        self.is_inclusion_list_empty = all(not r for r in [self.ruids,
                                                           self.phases,
                                                           self.tags,
                                                           self.levels])

    # def _not_int_list(self, lst):
    #    return [item for item in lst if not item.isdigit()]

    def _load_config(self, cfg: str, section: str) -> dict:  # pragma no cover
        raise NotImplementedError

    @staticmethod
    def _separate_values(data: str | Sequence[str]) -> tuple[Sequence[str], Sequence[str]]:
        """
        Separate included and excluded values based on sign pre-fixes from the given data.

        This method receives a list of data values and splits it into two lists:
        included values and excluded values. A data value is considered as 'excluded'
        if it starts with a '-' sign, while it is 'included' otherwise (including values without any prefix).

        Spaces and commas are treated as separators for the elements in the list. Non-string
        data elements will be converted to strings.
        
        Args:
            data (Sequence or str): A list of data values or a single string to separate.

        Returns:
            tuple: A tuple containing two lists; the first list consists of included values
                   (those not starting with '-'), and the second list consists of excluded values
                   (those starting with '-').

        Example:
            separate_values(["+apple", "-banana", "+cherry", "-date", "elderberry"])

            Should return (['apple', 'cherry', 'elderberry'], ['banana', 'date']) as
            "elderberry" doesn't start with a "-" sign, it is also included in the 'included' list.

        """
        data = data.replace(',', ' ').split() if isinstance(data, str) else data

        # Always make sure data elements are string
        data = [str(item) for item in data]

        # Separate included and excluded values note that + is optional
        included = [x.lstrip('+') for x in data if not x.startswith('-')]
        excluded = [x.lstrip('-') for x in data if x.startswith('-')]

        return included, excluded

    def expand_attributes(self, rc_data: dict):
        """
        Convert the data from the configuration dictionary into a form that
        is useful for processing in code.
        Args:
            rc_data: dictionary of attributes that should have all expected values

        Exception: If the levels list has an integer it throws an exception
        """

        self.ruids, self.ex_ruids = self._separate_values(rc_data.get('ruids', []))
        self.tags, self.ex_tags = self._separate_values(rc_data.get('tags', []))
        self.phases, self.ex_phases = self._separate_values(rc_data.get('phases', []))
        self.levels, self.ex_levels = self._separate_values(rc_data.get('levels', []))

    def does_match(self, ruid: str = "", tag: str = "", phase: str = "", level: str = "") -> bool:
        """
        Determines whether a given `ruid`/`tag`/`phase`/`level` matches any of the inclusions
        defined, and doesn't match any of the exclusions.

        With no inclusions defined, all values are included by default. Exclusion matching takes 
        precedent over inclusion matching.

        Args:
            ruid (str): The RUID string to check.
            tag (str): The tag string to check.
            phase (str): The phase string to check.
            level (int): The level integer to check.

        Returns:
            bool: True if it matches any inclusion and doesn't match any exclusion, False otherwise.
        """

        # This is sort of a hack levels must be integers, this makes any non integer level not match
        level = str(level)

        # This is a bit problematic because levels are ints not strs
        patterns: list[tuple[Sequence[str], str]] = [(self.ruids, ruid),
                                                     (self.tags, tag),
                                                     (self.levels, level),
                                                     (self.phases, phase)]
        ex_patterns: list[tuple[Sequence[str], str]] = [(self.ex_ruids, ruid),
                                                        (self.ex_tags, tag),
                                                        (self.ex_levels, level),
                                                        (self.ex_phases, phase)]

        # Check if any of the inputs match an inclusion pattern
        if not self.is_inclusion_list_empty:
            for pattern_list, attribute in patterns:
                if not attribute:
                    continue
                if pattern_list and not any(re.fullmatch(pat, attribute) for pat in pattern_list):
                    return False

        # Check if any of the inputs match an exclusion pattern
        for ex_pattern_list, attribute in ex_patterns:
            if not attribute:
                continue
            if any(re.fullmatch(exclusion, attribute) for exclusion in ex_pattern_list):
                return False

        return True
